from .ExperimentalValueFunction import ExperimentalValueFunction
from .MFValueFunction import MFValueFunction


class ICF(MFValueFunction):
    def __init__(self, var, user_var, item_var, stop_criteria, iterations,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.var = var
        self.user_var = user_var
        self.item_var = item_var
        self.stop_criteria = stop_criteria
        self.iterations = iterations

    def get_user_lambda(self):
        return self.var / self.user_var



from .ExperimentalValueFunction import ExperimentalValueFunction


class MFValueFunction(ExperimentalValueFunction):
    def __init__(self, num_lat, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_lat = num_lat




import numpy as np
from tqdm import tqdm
from . import ValueFunction, Entropy, Entropy0, MostPopular, LogPopEnt, ExperimentalValueFunction
import matplotlib.pyplot as plt
import scipy.stats
import os
import random


class EMostPopular(ExperimentalValueFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def top_emostpopular(self, items_exploitation, items_exploration):
        top_iids = []
        num_total_items = self.train_consumption_matrix.shape[1]
        for i in range(num_total_items):
            not_recommended = np.ones(num_total_items, dtype=bool)
            not_recommended[top_iids] = 0
            items_not_recommended = np.nonzero(not_recommended)[0]
            if self.epsilon < np.random.rand():
                best_item = items_not_recommended[np.argmax(
                    items_exploitation[items_not_recommended])]
            else:
                # best_item = random.choices(items_not_recommended,
                #                            weights=items_entropy[items_not_recommended]
                #                            ,k=1)[0]
                best_item = items_not_recommended[np.argmax(
                    items_exploration[items_not_recommended])]
            top_iids.append(best_item)
        return top_iids

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.items_entropy = Entropy.get_items_entropy(
            self.train_consumption_matrix)
        self.items_popularity = MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)

        self.top_iids = self.top_emostpopular(self.items_popularity,
                                              self.items_entropy)
        self.items_score = np.empty(len(self.top_iids))
        for i, iid in enumerate(reversed(self.top_iids)):
            self.items_score[iid] = i

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = self.items_score[candidate_items]
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        pass
        # top_iids = []
        # num_total_items = self.train_consumption_matrix.shape[1]
        # for i in range(num_total_items):
        #     not_recommended = np.ones(num_total_items,dtype=bool)
        #     not_recommended[top_iids] = 0
        #     items_not_recommended = np.nonzero(not_recommended)[0]
        #     if self.epsilon < np.random.rand():
        #         best_item = items_not_recommended[np.argmax(items_popularity[items_not_recommended])]
        #     else:
        #         # best_item = random.choices(items_not_recommended,
        #         #                            weights=items_entropy[items_not_recommended]
        #         #                            ,k=1)[0]
        #         best_item = items_not_recommended[np.argmax(items_entropy[items_not_recommended])]
        #     top_iids.append(best_item)

        # top_popularity_iids = list(reversed(np.argsort(items_entropy)))[:self.get_iterations()]
        # top_entropy_iids = list(reversed(np.argsort(items_popularity)))[:self.get_iterations()]

        # correlation = scipy.stats.pearsonr(items_entropy,items_popularity)[0]

        # # top_iids = list(reversed(np.argsort(items_popplusent)))[:self.get_iterations()]

        # fig, ax = plt.subplots()
        # ax.scatter(items_entropy,items_popularity,marker="D",color='darkblue')
        # ax.set_ylabel("Popularity")
        # ax.set_xlabel("Entropy")
        # ax.text(0.3, 0.9 , f'Correlation coefficient: {correlation:.2f}', color='k',
        #         ha='center', va='center',
        #         bbox=dict(facecolor='none', edgecolor='k', pad=10.0),
        #         transform = ax.transAxes)
        # for start, end, color in [(0,10,'green'),(10,20,'red'),(20,30,'darkred'),(30,40,'yellow'),(40,50,'orange')]:
        #     ax.scatter(items_entropy[top_iids[start:end]],items_popularity[top_iids[start:end]],marker='D',color=color)
        #     print("[%d,%d] sum(popularity)=%.2f sum(entropy)=%.2f"%(start,end,
        #                                                          np.sum(items_popularity[top_iids[start:end]]/np.max(items_popularity)),
        #                                                          np.sum(items_entropy[top_iids[start:end]]/np.max(items_entropy))))
        # fig.savefig(os.path.join(self.DIRS['img'],"corr_popent_"+self.get_id()+".png"))

        # num_total_users = len(uids)
        # for idx_uid in tqdm(range(num_total_users)):
        #     uid = uids[idx_uid]
        #     self.results[uid].extend(top_iids)
        # self.save_results()



import numpy as np
from tqdm import tqdm
from .ExperimentalValueFunction import ExperimentalValueFunction
from . import Entropy, MostPopular
import matplotlib.pyplot as plt
import os
import scipy.stats


class MostRepresentative(ExperimentalValueFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_items_representativeness(items_latent_factors):
        # return np.linalg.norm(items_latent_factors,axis=1,ord=np.inf)
        return np.sum(items_latent_factors**2, axis=1)

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        # uids = self.test_users
        self.items_representativeness = self.get_items_representativeness(
            items_latent_factors)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = self.items_representativeness[candidate_items]
        return items_score, None

        # top_iids = list(reversed(np.argsort(items_representativeness)))[:self.get_iterations()]

        # num_total_users = len(uids)
        # items_entropy = Entropy.get_items_entropy(self.train_consumption_matrix)
        # items_popularity = MostPopular.get_items_popularity(self.train_consumption_matrix,normalize=False)

        # correlation = scipy.stats.pearsonr(items_entropy,items_popularity)[0]
        # fig, ax = plt.subplots()
        # ax.scatter(items_entropy,items_popularity,marker="D",color='darkblue')
        # ax.set_ylabel("Popularity")
        # ax.set_xlabel("Entropy")
        # ax.text(0.3, 0.9 , f'Correlation coefficient: {correlation:.2f}', color='k',
        #         ha='center', va='center',
        #         bbox=dict(facecolor='none', edgecolor='k', pad=10.0),
        #         transform = ax.transAxes)
        # for start, end, color in [(0,10,'green'),(10,20,'red'),(20,30,'darkred'),(30,40,'yellow'),(40,50,'orange')]:
        #     ax.scatter(items_entropy[top_iids[start:end]],items_popularity[top_iids[start:end]],marker='D',color=color)
        # fig.savefig(os.path.join(self.DIRS['img'],"corr_popent_"+self.get_id()+".png"))

        # for idx_uid in tqdm(range(num_total_users)):
        #     uid = uids[idx_uid]
        #     self.results[uid].extend(top_iids)
        # self.save_results()



from .ICF import ICF
from .LinearICF import LinearICF
import numpy as np
from tqdm import tqdm
#import util
from threadpoolctl import threadpool_limits
import scipy.optimize
import ctypes
from collections import defaultdict
import joblib
import scipy
import mf
import value_functions


class GLM_UCB(LinearICF):
    """Generalized Linear Model Bandit-Upper Confidence Bound.
    
    It follows a similar process as Linear UCB based on the PMF formulation, but it also
    adds a sigmoid form in the exploitation step and makes a time-dependent exploration [1]_.

    References
    ----------
    .. [1] Zhao, Xiaoxue, Weinan Zhang, and Jun Wang. "Interactive collaborative filtering." 
       Proceedings of the 22nd ACM international conference on Information & Knowledge Management. 2013.
    """
    def __init__(self, c=1.0, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            c (float):
        """
        super().__init__(*args, **kwargs)
        self.c = c


    def sigmoid(self, x):
        return 1 / (1 + np.exp(-x))

    def p(self, x):
        return x

    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """ 
        train_dataset = observation
        super().reset(train_dataset)
        self.users_rec_rewards = defaultdict(list)
        self.users_rec_items_means = defaultdict(list)
        self.p_vals = dict()
        np.seterr(under="ignore")
        self.recent_predict = True
        self.t = 0

    def error_user_weight_function(self, p, u_rec_rewards, u_rec_items_means):
        return np.sum(
            np.array([(u_rec_rewards[t] - self.p(p.T @ u_rec_items_means[t])) *
                      u_rec_items_means[t]
                      for t in range(0, len(u_rec_items_means))]), 0)

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)
        
        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        A = self.As[uid]
        if len(self.users_rec_items_means[uid]) == 0:
            # self.p_vals[uid] = np.zeros(self.num_latent_factors)
            # self.p_vals[uid] = self.bs[uid]+1
            # print(self.bs[uid].dtype)
            self.p_vals[uid] = self.bs[uid]
            # self.p_vals[uid] = np.array(self.bs[uid],dtype=float)
        else:
            # self.p_vals[uid] = np.array(self.p_vals[uid],dtype=float)
            self.p_vals[uid] = scipy.optimize.root(
                self.error_user_weight_function, self.p_vals[uid],
                (self.users_rec_rewards[uid],
                 self.users_rec_items_means[uid])).x
        cov = np.linalg.inv(A) * self.var
        items_score = self.p(self.p_vals[uid][None,:] @ self.items_means[candidate_items].T) +\
            self.c * np.sqrt(np.log(self.t+1)) *\
            np.sqrt(np.sum(self.items_means[candidate_items].dot(cov) *\
                           self.items_means[candidate_items],axis=1))
        items_score = items_score.flatten()
        self.recent_predict = True
        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_mean = self.items_means[item, :]
        self.users_rec_rewards[uid].append(reward)
        self.users_rec_items_means[uid].append(max_item_mean)
        self.As[uid] += max_item_mean[:, None].dot(max_item_mean[None, :])
        if self.recent_predict:
            self.t += 1
            self.recent_predict = False


class GLM_UCBInit(GLM_UCB):
    def __init__(self, init, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init = init


    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)

        if self.init == 'entropy':
            items_entropy = value_functions.Entropy.get_items_entropy(
                self.train_consumption_matrix)
            self.items_bias = items_entropy
        elif self.init == 'popularity':
            items_popularity = value_functions.MostPopular.get_items_popularity(
                self.train_consumption_matrix, normalize=False)
            self.items_bias = items_popularity
        elif self.init == 'logpopent':
            items_entropy = value_functions.Entropy.get_items_entropy(
                self.train_consumption_matrix)
            items_popularity = value_functions.MostPopular.get_items_popularity(
                self.train_consumption_matrix, normalize=False)
            self.items_bias = value_functions.LogPopEnt.get_items_logpopent(
                items_popularity, items_entropy)
        elif self.init == 'rand_popularity':
            items_popularity = value_functions.MostPopular.get_items_popularity(
                self.train_consumption_matrix, normalize=False)
            items_popularity[np.argsort(items_popularity)[::-1][100:]] = 0
            self.items_bias = items_popularity
        elif self.init == 'random':
            self.items_bias = np.random.rand(
                self.train_dataset.num_total_items)

        self.items_bias = self.items_bias - np.min(self.items_bias)
        self.items_bias = self.items_bias / np.max(self.items_bias)

        assert (self.items_bias.min() >= 0
                and np.isclose(self.items_bias.max(), 1))

        res = scipy.optimize.minimize(
            lambda x, items_means, items_bias: np.sum(
                (items_bias - x @ items_means.T)**2),
            np.ones(self.num_latent_factors),
            args=(self.items_means, self.items_bias),
            method='BFGS',
        )
        self.initial_b = res.x

        print(
            np.corrcoef(self.items_bias,
                        self.initial_b @ self.items_means.T)[0, 1])

        self.bs = defaultdict(lambda: self.initial_b.copy())


class GLM_UCBEntropy(GLM_UCBInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init='entropy', *args, **kwargs)


class GLM_UCBPopularity(GLM_UCBInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init='popularity', *args, **kwargs)


class GLM_UCBRandPopularity(GLM_UCBInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init='rand_popularity', *args, **kwargs)


class GLM_UCBRandom(GLM_UCBInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init='random', *args, **kwargs)


class GLM_UCBLogPopEnt(GLM_UCBInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init='logpopent', *args, **kwargs)



from threadpoolctl import threadpool_limits
import copy
import numpy as np
from tqdm import tqdm
from .ExperimentalValueFunction import ExperimentalValueFunction
import matplotlib.pyplot as plt
import os
import scipy.sparse
import scipy.stats
from collections import defaultdict
import random
from .MFValueFunction import MFValueFunction
from tqdm import tqdm
from numba import njit, jit
import mf
import joblib
from .MostPopular import *


@njit
def _softmax(x):
    return np.exp(x - np.max(x)) / np.sum(np.exp(x - np.max(x)))


class _Particle:
    def __init__(self, num_users, num_items, num_lat):
        self.num_users = num_users
        self.num_items = num_items
        self.num_lat = num_lat
        self.alpha = 1
        self.beta = 1
        # self.lambda_ = np.ones(shape=(num_users,num_lat))
        # self.lambda_ = np.ones(shape=(num_lat,num_users))
        # self.lambda_ = np.ones(shape=(num_lat,num_users))
        self.lambda_ = np.ones(shape=(num_lat))
        # self.lambda_ = np.ones(shape=(num_lat))
        self.eta = np.ones(shape=(num_lat, num_items))
        self.mu = np.ones(shape=(num_items, num_lat))
        # self.Sigma = np.zeros(shape=(num_items,num_lat,num_lat))
        self.Sigma = np.array([np.identity(num_lat) for _ in range(num_items)])
        # self.sigma_n_2 = 1
        self.sigma_n_2 = scipy.stats.invgamma(self.alpha, self.beta).rvs()
        # self.p = np.random.dirichlet(self.lambda_,shape=(num_users,num_lat))
        self.p = np.array(
            [np.random.dirichlet(self.lambda_[:]) for uid in range(num_users)])
        # self.q = np.ones(shape=(num_items,num_lat))
        self.q = np.array([
            np.random.multivariate_normal(self.mu[i, :],
                                          self.sigma_n_2 * self.Sigma[i, :])
            for i in range(num_items)
        ])
        # self.Phi = np.ones(shape=(num_lat,num_items))
        self.Phi = np.array(
            [np.random.dirichlet(self.eta[i, :]) for i in range(num_lat)])

    # def p_expectations(self,uid,topic=None,reward=None):
    def p_expectations(self, uid, topic=None, reward=None):
        computed_sum = np.sum(self.lambda_[:])
        user_lambda = np.copy(self.lambda_[:])
        if reward != None:
            # user_lambda[topic] += reward
            user_lambda[topic] += reward
            computed_sum += reward
        return user_lambda / computed_sum

    def Phi_expectations(self, item, reward=None):
        # eta = np.copy(self.eta)
        computed_sum = np.sum(self.eta, axis=1)
        item_eta = np.copy(self.eta[:, item])
        if reward != None:
            item_eta += reward
            computed_sum += reward
        return item_eta / computed_sum

    def particle_weight(self, uid, item, reward):
        # print(self.p[uid],self.q[item])
        norm_val = scipy.stats.norm(self.p[uid] @ self.q[item],
                                    self.sigma_n_2).pdf(reward)
        return np.sum(norm_val * self.p_expectations(uid) *
                      self.Phi_expectations(item))

    def compute_theta(self, uid, item, reward, topic):
        return self.p_expectations(uid, reward=reward,
                                   topic=topic) * self.Phi_expectations(
                                       item, reward=reward)

    def select_z_topic(self, uid, item, reward):
        # topic = np.argmax(np.random.multinomial(1,self.p[uid]))
        # topic = None
        topic = np.argmax(
            np.random.multinomial(1, [1 / self.num_lat] * self.num_lat))
        theta = self.compute_theta(uid, item, reward, topic)
        # print(np.sum(theta))
        # theta = _softmax(theta)
        # print(np.sum(theta))
        theta = theta / np.sum(theta)
        topic = np.argmax(np.random.multinomial(1, theta))
        # if uid == 57 or uid == 98 or uid == 47:
        # if uid == 47:
        # print(uid,item,theta,topic)
        return topic

    def update_parameters(self, uid, item, reward, topic):
        new_Sigma = np.linalg.inv(
            np.linalg.inv(self.Sigma[item]) +
            self.p[uid][:, None] @ self.p[uid][None, :])
        new_mu = new_Sigma @ (np.linalg.inv(self.Sigma[item]) @ self.mu[item] +
                              self.p[uid] * reward)
        new_alpha = self.alpha + 1 / 2
        new_beta = self.beta + 1 / 2 * (
            self.mu[item].T @ np.linalg.inv(self.Sigma[item]) @ self.mu[item] +
            reward * reward - new_mu.T @ np.linalg.inv(new_Sigma) @ new_mu)
        # new_lambda_k = self.lambda_[topic]+reward
        new_lambda_k = self.lambda_[topic] + reward
        new_eta_k = self.eta[topic, item] + reward

        self.Sigma[item] = new_Sigma
        self.mu[item] = new_mu
        self.alpha = new_alpha
        self.beta = new_beta
        self.lambda_[topic] = new_lambda_k
        # self.eta[:,item] = new_eta_k
        self.eta[topic, item] = new_eta_k

    def sample_random_variables(self, uid, item, topic):
        self.sigma_n_2 = scipy.stats.invgamma(self.alpha, self.beta).rvs()
        self.q[item] = np.random.multivariate_normal(
            self.mu[item], self.sigma_n_2 * self.Sigma[item])
        self.p[uid] = np.random.dirichlet(self.lambda_[:])
        self.Phi[topic] = np.random.dirichlet(self.eta[topic])


class ICTRTS(MFValueFunction):
    """Interactive Collaborative Topic Regression with Thompson Sampling.
    
    It is an interactive collaborative topic regression model that utilizes the TS
    bandit algorithm and controls the items dependency by a particle learning strategy [1]_.

    References
    ----------
    .. [1] Wang, Qing, et al. "Online interactive collaborative filtering using multi-armed 
       bandit with dependent arms." IEEE Transactions on Knowledge and Data Engineering 31.8 (2018): 1569-1580.   
    """
    def __init__(self, num_particles, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            num_particles (int):
        """
        super().__init__(*args, **kwargs)
        self.num_particles = num_particles


    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """ 
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))

        self.num_total_items = self.train_dataset.num_total_items
        self.num_total_users = self.train_dataset.num_total_users
        # self.items_popularity = MostPopular.get_items_popularity(self.train_consumption_matrix, normalize=False)
        # self.particles = [_Particle(self.num_total_users,self.num_total_items,self.num_lat) for i in range(self.num_particles)]
        particle = _Particle(self.num_total_users, self.num_total_items,
                             self.num_lat)
        for i in tqdm(range(len(self.train_dataset.data))):
            uid = int(self.train_dataset.data[i, 0])
            item = int(self.train_dataset.data[i, 1])
            reward = self.train_dataset.data[i, 2]
            # reward = int(reward>=4)
            topic = particle.select_z_topic(uid, item, reward)
            particle.update_parameters(uid, item, reward, topic)
            particle.sample_random_variables(uid, item, topic)
            # self.update(uid,item,reward,None)
            # if i > 10000:
            # break

        self.particles = [
            copy.deepcopy(particle) for _ in range(self.num_particles)
        ]

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)
        
        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        # items_pool_size = max(int(len(candidate_items)*0.1),num_req_items)
        # items_pool_size = max(int(len(candidate_items)*0.1),num_req_items)
        # items_pool_size = max(min(len(candidate_items),100),num_req_items)
        # candidate_items = candidate_items[np.argsort(self.items_popularity[candidate_items])[::-1][:items_pool_size]]
        items_score = np.zeros(len(candidate_items))
        for particle in self.particles:
            # print(particle.p[uid].shape,particle.q[candidate_items].shape)
            items_score += particle.q[candidate_items, :] @ particle.p[uid, :]

        items_score /= self.num_particles
        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        # if reward > 0:
        # self.items_popularity[item] += 1
        # reward = int(reward>=4)
        # for particle in self.particles:
        # with np.errstate(under='ignore'):
        weights = [
            particle.particle_weight(uid, item, reward)
            for particle in self.particles
        ]
        weights = np.array(weights)
        weights = _softmax(weights)
        # copy.deepcopy
        ds = np.random.choice(range(self.num_particles),
                              p=weights,
                              size=self.num_particles)
        new_particles = []
        for i in range(self.num_particles):
            new_particles.append(copy.deepcopy(self.particles[ds[i]]))
        self.particles = new_particles
        for particle in self.particles:
            topic = particle.select_z_topic(uid, item, reward)
            particle.update_parameters(uid, item, reward, topic)
            particle.sample_random_variables(uid, item, topic)



from .ExperimentalValueFunction import ExperimentalValueFunction
import numpy as np
from tqdm import tqdm
#import util
from threadpoolctl import threadpool_limits
import ctypes
import functools
from .MFValueFunction import MFValueFunction


class TinUCB(MFValueFunction):
    def __init__(self, alpha=0.2, zeta=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if alpha != None:
            self.alpha = alpha
        elif zeta != None:
            self.alpha = 1 + np.sqrt(np.log(2 / zeta) / 2)


    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items
        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights
        self.I = np.eye(len(self.items_weights[0]))
        bs = defaultdict(lambda: np.zeros(self.num_latent_factors))
        tmp_I = self.I
        As = defaultdict(lambda: tmp_I.copy())
        self.users_A_sums_history = defaultdict(lambda: [tmp_I.copy()])

    def g(self, x):
        return x

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = bs[uid]
        A = As[uid]
        A_weights = [
            self.g(i + 1) / self.g(len(self.users_A_sums_history[uid]))
            for i in range(len(A_sums_history))
        ]

        A = functools.reduce(
            lambda a, b: a + b,
            map(lambda x, w: x * w, zip(A_sums_history, A_weights)))

        mean = np.dot(np.linalg.inv(A), b)
        items_score = mean @ self.items_weights[candidate_items].T+\
            self.alpha*np.sqrt(np.sum(self.items_weights[candidate_items].dot(np.linalg.inv(A)) * self.items_weights[candidate_items],axis=1))
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        b = bs[uid]
        A = As[uid]
        max_item_latent_factors = self.items_weights[item]
        self.users_A_sums_history[uid].append(
            max_item_latent_factors[:, None].dot(
                max_item_latent_factors[None, :]))
        A += max_item_latent_factors[:, None].dot(
            max_item_latent_factors[None, :])
        b += reward * max_item_latent_factors



from numpy.lib.npyio import save
from irec import value_functions
from irec import action_selection_policies
from irec.CandidateActions import CandidateActions
from irec.CandidateAction import CandidateAction, UICandidateAction
import numpy as np
from typing import List, Any

"""Recommender System Agent.

An agent can be defined as the Recommendation System (RS) itself. At each time step,
the Agent requires a recommendation (that is, performs an action) given a context or
state of the RS Environment. After offering a recommendation, either an item or a list of items,
the Agent will receive a reward from the Environment. In the RS field, rewards are usually
a scalar value representative of the user's explicit/implicit feedback.
"""


class Agent:
    """Agent.
    The agent's goal is to learn a policy, a strategy that indicates the correct action to be taken 
    in each state/situation, in order to maximize the total reward collected throughout the interaction.

    Every Agent has a value function, whose main task is to estimate the usefulness of a 
    recommendation for a user. From an RL perspective, in addition to an estimate of the 
    immediate utility/reward of an action, the agent aims to learn from experience the long-term 
    value of an action. In short, the value function incorporates the agent's goals, quantifying
    the expected long-term consequences of decisions.

    In addition, the action selection policy uses the information provided by the value function.
    to select the next action to be performed.
    """

    def __init__(
        self,
        value_function: value_functions.ValueFunction,
        action_selection_policy: action_selection_policies.ActionSelectionPolicy,
        name: str,
        *args,
        **kwargs
    ):

        """__init__.

        Args:
            value_function (value_functions.ValueFunction): value_function
            action_selection_policy (action_selection_policies.ActionSelectionPolicy): action_selection_policy
            name (str): name
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)
        self.value_function: value_functions.ValueFunction = value_function
        self.action_selection_policy = action_selection_policy
        self.name = name

    def act(self, candidate_actions: CandidateActions, actions_num: int):
        """act.

        An action is a recommendation, which will be made about the items available to a particular target user.

        Args:
            candidate_actions (CandidateActions): candidate_actions
            actions_num (int): actions_num
        """
        raise NotImplementedError

    def observe(self, observation: Any, action: CandidateAction, reward: float, info: dict):
        """observe.

        After each action, the agent receives a reward, and determines what kind of information the agent 
        can observe to update its value function and its selection policy.

        Args:
            observation (Any): observation
            action (CandidateAction): action
            reward (float): reward
            info (dict): info
        """

        raise NotImplementedError

    def reset(self, observation: Any):
        """reset.

        Reset agent parameters.

        Args:
            observation (Any): observation
        """

        raise NotImplementedError


class SimpleAgent(Agent):
    """SimpleAgent."""

    def __init__(self, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)

    def act(self, candidate_actions: CandidateActions, actions_num: int):
        """act.

        Args:
            candidate_actions (CandidateActions): candidate_actions
            actions_num (int): actions_num
        """
        action_estimates, vf_info = self.value_function.action_estimates(
            candidate_actions
        )
        actions, asp_info = self.action_selection_policy.select_actions(
            candidate_actions, action_estimates, actions_num
        )
        # actions = (candidate_actions[0],candidate_actions[1][actions_indexes])
        return actions, {"vf_info": vf_info, "asp_info": asp_info}

    def observe(
        self, observation: Any, action: UICandidateAction, reward: float, info: dict
    ):
        """observe.
        
        Args:
            observation (Any): observation
            action (UICandidateAction): action
            reward (float): reward
            info (dict): info
        """
        vf_info = info.get("vf_info", None)
        asp_info = info.get("asp_info", None)
        self.value_function.update(observation, action, reward, vf_info)
        self.action_selection_policy.update(observation, action, reward, asp_info)
        return None

    def reset(self, observation: Any):
        """reset.

        Args:
            observation (Any): observation
        """
        if observation != None:
            self.value_function.reset(observation)
            self.action_selection_policy.reset(observation)
            return None
        else:
            raise NotImplementedError


class SimpleEnsembleAgent(Agent):
    """SimpleEnsembleAgent."""

    def __init__(
        self,
        agents: List[Agent],
        use_name_meta_actions: bool = True,
        save_meta_actions: bool = True,
        *args,
        **kwargs
    ):
        """__init__.

        Args:
            agents (List[Agent]): agents
            use_name_meta_actions (bool): use_name_meta_actions
            save_meta_actions (bool): save_meta_actions
            args:
            kwargs:
        """

        super().__init__(*args, **kwargs)
        # self.ensemble_method_vf = ensemble_method_vf
        self.agents = agents
        self.use_name_meta_actions = use_name_meta_actions
        if self.use_name_meta_actions:
            self.ensemble_candidate_actions = np.array([i.name for i in self.agents])
            self.actions_name_object_map = {
                act_name: self.agents[i]
                for i, act_name in enumerate(self.ensemble_candidate_actions)
            }
        else:
            self.ensemble_candidate_actions = np.array(list(range(len(self.agents))))
        # self.ensemble_candidate_actions = []
        self.default_actions_num = 1
        self.save_meta_actions = save_meta_actions

    def act(self, candidate_actions: CandidateActions, actions_num: int):
        """act.

        Args:
            candidate_actions (CandidateActions): candidate_actions
            actions_num (int): actions_num
        """
        info = {}
        meta_action_estimates, meta_vf_info = self.value_function.action_estimates(
            self.ensemble_candidate_actions
        )
        # print(m,meta_action_estimates)
        # info = {'meta_action_estimates': meta_action_estimates}
        meta_actions, meta_asp_info = self.action_selection_policy.select_actions(
            self.ensemble_candidate_actions,
            meta_action_estimates,
            self.default_actions_num,
        )
        meta_action = meta_actions[0]
        # print(meta_actions)
        # actions = (candidate_actions[0],candidate_actions[1][actions_indexes])
        if self.use_name_meta_actions:
            selected_agent = self.actions_name_object_map[meta_action]
        else:
            selected_agent = self.agents[meta_action]
        selected_agent_actions, selected_agent_info = selected_agent.act(
            candidate_actions, actions_num
        )
        if meta_vf_info != None:
            info["vf_info"] = meta_vf_info
        if meta_asp_info != None:
            info["asp_info"] = meta_asp_info
        if selected_agent_info != None:
            info["selected_agent_info"] = selected_agent_info
        if self.save_meta_actions:
            if self.use_name_meta_actions:
                info["meta_action_name"] = meta_action
                info["mai"] = meta_action
            else:
                info["meta_action_name"] = self.agents[meta_action].name
                info["mai"] = meta_action
        if info == dict():
            info = None
        return selected_agent_actions, info

    def observe(self, observation: Any, action: UICandidateAction, reward: float, info: dict):
        """observe.

        Args:
            observation (Any): observation
            action (UICandidateAction): action
            reward (float): reward
            info (dict): info
        """
        vf_info = info.get("vf_info", None)
        asp_info = info.get("asp_info", None)
        self.value_function.update(observation, info["mai"], reward, vf_info)
        self.action_selection_policy.update(observation, info["mai"], reward, asp_info)
        for agent in self.agents:
            agent.observe(observation, action, reward, info)
        return None

    def reset(self, observation: Any):
        """reset.

        Args:
            observation (Any): observation
        """
        if observation != None:
            self.value_function.reset(observation)
            self.action_selection_policy.reset(observation)
            for agent in self.agents:
                agent.reset(observation)
            return None
        else:
            raise NotImplementedError



from copy import copy
import numpy as np
import pandas as pd
import random


class Splitter:
    pass


class TrainTestConsumption(Splitter):
    def __init__(self,
                 train_size=0.8,
                 test_consumes=1,
                 crono=False,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.train_size = 0.8
        self.test_consumes = test_consumes
        self.crono = crono

    def process(self, dataset):
        data = dataset.data
        num_users = len(np.unique(data[:, 0]))
        num_train_users = round(num_users * (self.train_size))
        num_test_users = int(num_users - num_train_users)
        data_df = pd.DataFrame(data)
        users_items_consumed = data_df.groupby(0).count().iloc[:, 0]
        test_candidate_users = list(users_items_consumed[
            users_items_consumed >= self.test_consumes].to_dict().keys())
        if self.crono:
            users_start_time = data_df.groupby(0).min()[3].to_numpy()
            test_uids = np.array(
                list(test_candidate_users[list(
                    reversed(np.argsort(
                        users_start_time[test_candidate_users])))])
                [:num_test_users])
        else:
            test_uids = np.array(
                random.sample(test_candidate_users, k=num_test_users))
            # print(test_uids)
        train_uids = np.array(list(set(range(num_users)) - set(test_uids)))

        data_isin_test_uids = np.isin(data[:, 0], test_uids)

        train_dataset = copy(dataset)
        train_dataset.data = data[~data_isin_test_uids, :]
        dataset.update_from_data()
        test_dataset = copy(dataset)
        test_dataset.data = data[data_isin_test_uids, :]
        dataset.update_from_data()
        print("Test shape:", test_dataset.data.shape)
        print("Train shape:", train_dataset.data.shape)
        return train_dataset, test_dataset


class TRTETrainValidation(Splitter):
    def __init__(self,
                 train_size=0.8,
                 test_consumes=1,
                 crono=False,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.train_size = 0.8
        self.test_consumes = test_consumes
        self.crono = crono

    def process(self, train_dataset, test_dataset):
        ttc = TrainTestConsumption(self.train_size, self.test_consumes,
                                   self.crono)
        train_dataset, test_dataset = ttc.process(train_dataset)
        return train_dataset, test_dataset



from os.path import dirname, realpath, sep, pardir
import sys, os

sys.path.append(dirname(realpath(__file__)) + sep + pardir)

import numpy as np
import scipy.sparse
from numba import jit, prange

# from .. import irec.utils
import metrics


@jit(nopython=True, parallel=True)
def _predict_sparse(users_weights, items_weights, users_items):
    n = len(users_items[0])
    results = np.zeros(n)
    for i in prange(n):
        uid = users_items[0][i]
        iid = users_items[1][i]
        results[i] = users_weights[uid] @ items_weights[iid]
    return results


class MF:
    """MF."""

    def __init__(self, num_lat=10, *args, **kwargs):
        """__init__.

        Args:
            num_lat:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)
        self.num_lat = num_lat

    def normalize_matrix(self, matrix):
        return matrix / np.max(matrix)

    def fit(self):
        """fit."""
        pass

    def predict_sparse(self, users_items):
        return _predict_sparse(self.users_weights, self.items_weights, users_items)

    def predict(self, X):
        """predict.

        Args:
            X:
        """
        if isinstance(X, scipy.sparse.spmatrix):
            observed_ui = (X.tocoo().row, X.tocoo().col)
            X = observed_ui
        return self.predict_sparse(X)

    def score(self, X):
        """score.

        Args:
            X:
        """
        return metrics.rmse(X.data, self.predict(X))



import numpy as np
from tqdm import tqdm
#import util
from threadpoolctl import threadpool_limits
import ctypes
import scipy.spatial
import matplotlib.pyplot as plt
import os
import pickle
import mf
from .MFValueFunction import MFValueFunction
from collections import defaultdict
import value_functions


class OurMethod1(MFValueFunction):
    def __init__(self,
                 alpha=1.0,
                 stop=None,
                 weight_method='change',
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.alpha = alpha
        self.weight_method = weight_method
        self.stop = stop


    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items

        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights
        self.num_latent_factors = len(mf_model.items_weights[0])

        # self.items_weights = items_weights
        # num_total_users = len(uids)

        items_entropy = value_functions.Entropy.get_items_entropy(
            self.train_consumption_matrix)
        items_popularity = value_functions.MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)
        # self.items_bias = value_functions.PPELPE.get_items_ppelpe(items_popularity,items_entropy)
        self.items_bias = value_functions.LogPopEnt.get_items_logpopent(
            items_popularity, items_entropy)

        assert (self.items_bias.min() >= 0 and self.items_bias.max() == 1)

        self.I = np.eye(len(self.items_weights[0]))
        self.bs = defaultdict(lambda: np.zeros(self.num_latent_factors))
        self.As = defaultdict(lambda: self.I.copy())

        self.users_latent_factors_history = defaultdict(
            lambda: np.empty(shape=(0, self.num_latent_factors)))
        self.users_num_correct_items_history = defaultdict(lambda: [0])
        self.users_similarity_score = defaultdict(lambda: [0])
        self.users_distance_history = defaultdict(lambda: [])
        self.users_global_model_weights = defaultdict(lambda: [])

        # user_latent_factors_history = np.empty(shape=(0, num_lat))
        # num_correct_items_history = [0]
        # similarity_score = [0]
        # distance_history = []
        # global_model_weights = []

    def get_global_model_weight(self, user_latent_factors_history,
                                num_correct_items_history, distance_history):
        if self.weight_method == 'stop':
            b = self.stop
            a = min(np.sum(num_correct_items_history), b)
            return (1 - np.round(pow(2, a) / pow(2, b), 3))
        elif self.weight_method == 'change':
            if len(user_latent_factors_history) == 0:
                return 1
            times_with_reward = np.nonzero(num_correct_items_history)[0]
            if len(times_with_reward) < 2:
                return 1
            # print(user_latent_factors_history[times_with_reward][-1])
            # print(user_latent_factors_history[times_with_reward][-2])
            res = scipy.spatial.distance.cosine(
                user_latent_factors_history[times_with_reward][-1],
                user_latent_factors_history[times_with_reward][-2])
            distance_history.append(res)
            res = res / np.max(distance_history)
            # res = (res+1)/2
            return res
        else:
            raise RuntimeError

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]

        self.users_latent_factors_history[uid]
        num_correct_items_history = self.users_num_correct_items_history[uid]
        similarity_score = self.users_similarity_score[uid]
        distance_history = self.users_distance_history[uid]
        global_model_weights = self.users_global_model_weights[uid]

        user_latent_factors = np.dot(np.linalg.inv(A), b)
        self.users_latent_factors_history[uid] = np.vstack(
            [self.users_latent_factors_history[uid], user_latent_factors])
        global_model_weight = self.get_global_model_weight(
            self.users_latent_factors_history[uid], num_correct_items_history,
            distance_history)
        global_model_weights.append(global_model_weight)
        items_uncertainty = np.sqrt(
            np.sum(self.items_weights[candidate_items].dot(np.linalg.inv(A)) *
                   self.items_weights[candidate_items],
                   axis=1))
        items_user_similarity = user_latent_factors @ self.items_weights[
            candidate_items].T
        user_model_items_score = items_user_similarity + self.alpha * items_uncertainty
        global_model_items_score = self.items_bias[candidate_items]
        user_model_items_score_min = np.min(user_model_items_score)
        user_model_items_score_max = np.max(user_model_items_score)
        if user_model_items_score_max - user_model_items_score_min != 0:
            global_model_items_score = global_model_items_score * (
                user_model_items_score_max -
                user_model_items_score_min) + user_model_items_score_min

        items_score = (
            1 - global_model_weight
        ) * user_model_items_score + global_model_weight * global_model_items_score
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_latent_factors = self.items_weights[item]
        b = self.bs[uid]
        A = self.As[uid]
        A += max_item_latent_factors[:, None].dot(
            max_item_latent_factors[None, :])
        b += reward * max_item_latent_factors
        if reward > min(self.train_dataset.rate_domain):
            num_correct_items_history = self.users_num_correct_items_history[
                uid]
            num_correct_items_history[-1] += 1



from .ExperimentalValueFunction import *
import numpy as np
import random
from tqdm import tqdm
from threadpoolctl import threadpool_limits
import mf
import ctypes
from .MFValueFunction import MFValueFunction


class LinEGreedy(MFValueFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.epsilon = epsilon
        #

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items
        mf_model = mf.SVD()
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights

        self.bs = defaultdict(lambda: np.ones(self.num_lat))
        self.As = defaultdict(lambda: np.eye(self.num_lat))

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]

        mean = np.dot(np.linalg.inv(A), b)
        items_score = mean @ self.items_weights[candidate_items].T

        # rand = np.random.rand(min(num_req_items, len(candidate_items)))
        # rand = self.epsilon > rand

        # cnz = np.count_nonzero(rand)
        # if cnz == min(num_req_items, len(candidate_items)):
        # items_score = np.random.rand(len(candidate_items))
        # else:
        # items_score = mean @ self.items_weights[candidate_items].T
        # randind = random.sample(list(range(len(candidate_items))),
        # k=np.count_nonzero(rand))
        # items_score[randind] = np.inf
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_mean = self.items_weights[item]
        self.As[uid] += max_item_mean[:, None].dot(max_item_mean[None, :])
        self.bs[uid] += reward * max_item_mean



import numpy as np
from tqdm import tqdm
from . import ValueFunction, Entropy, MostPopular, LogPopEnt, ExperimentalValueFunction
import matplotlib.pyplot as plt
import scipy.stats
import os


class PopPlusEnt(ExperimentalValueFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_items_popplusent(items_popularity, items_entropy, log=False):
        if not log:
            items_popplusent = items_entropy / np.max(
                items_entropy) + items_popularity / np.max(items_popularity)
        else:
            items_popplusent = items_entropy / np.max(
                items_entropy) + np.ma.log(items_popularity).filled(
                    0) / np.max(np.ma.log(items_popularity).filled(0))
        return items_popplusent / np.max(items_popplusent)

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))

        items_entropy = Entropy.get_items_entropy(
            self.train_consumption_matrix)
        items_popularity = MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)
        self.items_popplusent = PopPlusEnt.get_items_popplusent(
            items_popularity, items_entropy)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = self.items_popplusent[candidate_items]
        return items_score, None

        # correlation = scipy.stats.pearsonr(items_entropy,items_popularity)[0]

        # top_iids = list(reversed(np.argsort(items_popplusent)))[:self.get_iterations()]
        # fig, ax = plt.subplots()
        # ax.hist(items_entropy,color='k')
        # ax.set_xlabel("$Pop + Ent$")
        # ax.set_ylabel("#Items")
        # fig.savefig(os.path.join(self.DIRS['img'],"popplusent_"+self.get_id()+".png"))

        # fig, ax = plt.subplots()
        # ax.scatter(items_entropy,items_popularity,marker="D",color='darkblue')
        # ax.set_ylabel("Popularity")
        # ax.set_xlabel("Entropy")
        # ax.text(0.3, 0.9 , f'Correlation coefficient: {correlation:.2f}', color='k',
        #         ha='center', va='center',
        #         bbox=dict(facecolor='none', edgecolor='k', pad=10.0),
        #         transform = ax.transAxes)
        # for start, end, color in [(0,10,'green'),(10,20,'red'),(20,30,'darkred'),(30,40,'yellow'),(40,50,'orange')]:
        #     print(top_iids[start:end])
        #     ax.scatter(items_entropy[top_iids[start:end]],items_popularity[top_iids[start:end]],marker='D',color=color)
        #     print("[%d,%d] sum(popularity)=%.2f sum(entropy)=%.2f"%(start,end,
        #                                                          np.sum(items_popularity[top_iids[start:end]]/np.max(items_popularity)),
        #                                                          np.sum(items_entropy[top_iids[start:end]]/np.max(items_entropy))))
        # fig.savefig(os.path.join(self.DIRS['img'],"corr_popent_"+self.get_id()+".png"))

        # num_total_users = len(uids)
        # for idx_uid in tqdm(range(num_total_users)):
        #     uid = uids[idx_uid]
        #     self.results[uid].extend(top_iids)
        # self.save_results()



import numpy as np
from tqdm import tqdm, trange
from .ExperimentalValueFunction import ExperimentalValueFunction
import os
import random
import scipy.stats
from collections import defaultdict


class ThompsonSampling(ExperimentalValueFunction):
    """Thompson Sampling.
    
    A basic item-oriented bandit algorithm that follows a Gaussian distribution
    of items and users to perform the prediction rule based on their samples [1]_.


    References
    ----------
    .. [1] Chapelle, Olivier, and Lihong Li. "An empirical evaluation of thompson sampling."
        Advances in neural information processing systems 24 (2011): 2249-2257.   
    """
    def __init__(self, alpha_0, beta_0, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            alpha_0:
            beta_0:
        """
        super().__init__(*args, **kwargs)
        self.alpha_0 = alpha_0
        self.beta_0 = beta_0


    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items

        self.alphas = np.ones(self.num_total_items) * self.alpha_0
        self.betas = np.ones(self.num_total_items) * self.beta_0

        for i in range(self.train_dataset.data.shape[0]):
            uid = int(self.train_dataset.data[i, 0])
            item = int(self.train_dataset.data[i, 1])
            reward = self.train_dataset.data[i, 2]
            # self.update(uid, item, reward, None)
            self.update(None, (uid, item), reward, None)

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)

        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = np.random.beta(self.alphas[candidate_items],
                                     self.betas[candidate_items])
        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        reward = 1 if (reward >= self.train_dataset.mean_rating) else 0
        self.alphas[item] += reward
        self.betas[item] += 1 - reward



from .ICF import ICF
import numpy as np
from tqdm import tqdm
from threadpoolctl import threadpool_limits
import ctypes
from collections import defaultdict
import joblib
import scipy
import mf


class LinearICF(ICF):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (
                self.train_dataset.data[:, 2],
                (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1]),
            ),
            (self.train_dataset.num_total_users, self.train_dataset.num_total_items),
        )
        self.num_total_items = self.train_dataset.num_total_items
        mf_model = mf.ICFPMFS(
            self.iterations,
            self.var,
            self.user_var,
            self.item_var,
            self.stop_criteria,
            num_lat=self.num_lat,
        )
        mf_model.fit(self.train_consumption_matrix)

        self.items_means = mf_model.items_means

        self.num_latent_factors = len(self.items_means[0])

        self.I = np.eye(self.num_latent_factors)
        self.bs = defaultdict(lambda: np.zeros(self.num_latent_factors))
        self.As = defaultdict(lambda: self.get_user_lambda() * self.I)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]
        mean = np.dot(np.linalg.inv(A), b)
        cov = np.linalg.inv(A) * self.var

        items_score = mean @ self.items_means[candidate_items].T

        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_mean = self.items_means[item]
        b = self.bs[uid]
        A = self.As[uid]
        A += max_item_mean[:, None].dot(max_item_mean[None, :])
        b += reward * max_item_mean



from .ICF import ICF
import numpy as np
from tqdm import tqdm
from threadpoolctl import threadpool_limits
import ctypes
from collections import defaultdict
import joblib
import scipy
import mf
from .LinearICF import LinearICF


class LinearUCB(LinearICF):
    """LinearUCB.
    
    An adaptation of the original LinUCB (Lihong Li et al. 2010) to measure
    the latent dimensions by a PMF formulation [1]_.

    References
    ----------
    .. [1] Zhao, Xiaoxue, Weinan Zhang, and Jun Wang. "Interactive collaborative filtering." 
       Proceedings of the 22nd ACM international conference on Information & Knowledge Management. 2013.
    """
    def __init__(self, alpha, zeta=None, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            alpha:
        """
        super().__init__(*args, **kwargs)
        if alpha != None:
            self.alpha = alpha
        elif zeta != None:
            self.alpha = 1 + np.sqrt(np.log(2 / zeta) / 2)

    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """
        train_dataset = observation
        super().reset(train_dataset)

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)

        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]
        mean = np.dot(np.linalg.inv(A), b)
        cov = np.linalg.inv(A) * self.var

        items_score  = mean @ self.items_means[candidate_items].T+\
            self.alpha*np.sqrt(np.sum(self.items_means[candidate_items].dot(cov) * self.items_means[candidate_items],axis=1))

        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        return super().update(observation, action, reward, info)



import numpy as np
import scipy
import scipy.stats
import random
from threadpoolctl import threadpool_limits
import ctypes

from irec.utils.utils import run_parallel
from mf import MF
from numba import jit


@jit(nopython=True)
def _svdplusplus(data, indptr, indices, n_u, num_users, num_items, num_lat,
                 learn_rate, delta, delta_bias, bias_learn_rate, iterations,
                 stop_criteria, init_mean, init_std):
    # print(n_u)
    num_r = len(data)
    r_mean = np.mean(data)
    r_std = np.std(data)
    b_u = np.zeros(num_users)
    b_i = np.zeros(num_items)
    p = np.random.normal(init_mean, init_std, (num_users, num_lat))
    q = np.random.normal(init_mean, init_std, (num_items, num_lat))

    y = np.random.normal(init_mean, init_std, (num_items, num_lat))

    rmse_old = 0.0
    for iteration in range(iterations):
        error = 0.0
        for row in range(len(indptr) - 1):
            for i in range(indptr[row], indptr[row + 1]):
                r_ui = data[i]
                uid = row
                iid = indices[i]
                y_sum = np.zeros(num_lat)
                for j in range(indptr[row], indptr[row + 1]):
                    _iid = indices[j]
                    y_sum += y[_iid]
                y_sum = y_sum / n_u[uid]

                p_u = p[uid] + y_sum

                predicted_r = r_mean + b_u[uid] + b_i[iid] + (p_u @ q[iid])
                e_ui = r_ui - predicted_r
                # raise SystemExit
                # print(e_ui)
                # print(len(e_ui))
                # print(len(e_ui))
                # print(e_ui)
                error += e_ui**2

                b_u[uid] += bias_learn_rate * (e_ui - delta_bias * b_u[uid])
                b_i[iid] += bias_learn_rate * (e_ui - delta_bias * b_i[iid])

                normalized_e_ui = e_ui / n_u[uid]
                p[uid] += learn_rate * (normalized_e_ui * q[iid] -
                                        delta * p[uid])
                q[iid] += learn_rate * (normalized_e_ui * p_u - delta * q[iid])
        rmse = error / num_r
        print(iteration + 1, 'RMSE:', rmse)

        if np.fabs(rmse - rmse_old) <= stop_criteria:
            break
        else:
            rmse_old = rmse
    return b_u, b_i, p, q, y


class SVDPlusPlus(MF):
    def __init__(self,
                 iterations=50,
                 learn_rate=0.05,
                 delta=0.015,
                 delta_bias=0.002,
                 bias_learn_rate=0.005,
                 stop_criteria=0.009,
                 init_mean=0,
                 init_std=0.1,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.iterations = iterations
        self.learn_rate = learn_rate
        self.delta = delta
        self.delta_bias = delta_bias
        self.bias_learn_rate = bias_learn_rate
        self.stop_criteria = stop_criteria
        self.init_mean = init_mean
        self.init_std = init_std

    def fit(self, training_matrix):
        super().fit()
        num_users = training_matrix.shape[0]
        num_items = training_matrix.shape[1]
        n_u = np.array([np.sqrt(len(i.data)) for i in training_matrix])
        self.r_mean = np.mean(training_matrix.data)
        self.b_u, self.b_i, self.p, self.q, self.y = _svdplusplus(
            training_matrix.data, training_matrix.indptr,
            training_matrix.indices, n_u, num_users, num_items, self.num_lat,
            self.learn_rate, self.delta, self.delta_bias_bias,
            self.bias_learn_rate, self.iterations, self.stop_criteria,
            self.init_mean, self.init_std)



import numpy as np
from tqdm import tqdm
#import util
from threadpoolctl import threadpool_limits
import ctypes
import scipy.stats
import scipy.spatial
import matplotlib.pyplot as plt
import os
import pickle
import sklearn
import scipy.optimize
import scipy
import mf
from collections import defaultdict
from .MFValueFunction import MFValueFunction
from value_functions.Entropy import Entropy
from value_functions.MostPopular import MostPopular
from value_functions.LogPopEnt import LogPopEnt


def _prediction_rule(A, b, items_weights, alpha):
    user_latent_factors = np.dot(np.linalg.inv(A), b)
    items_uncertainty = np.sqrt(
        np.sum(items_weights.dot(np.linalg.inv(A)) * items_weights, axis=1))
    items_user_similarity = user_latent_factors @ items_weights.T
    weighted_items_uncertainty = alpha * items_uncertainty
    user_model_items_score = items_user_similarity + weighted_items_uncertainty
    return user_model_items_score, items_user_similarity,weighted_items_uncertainty


class OurMethod2(MFValueFunction):
    def __init__(self, alpha, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alpha = alpha


    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items

        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights
        self.num_latent_factors = len(self.items_weights[0])

        items_entropy = Entropy.get_items_entropy(
            self.train_consumption_matrix)
        items_popularity = MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)
        # self.items_bias = value_functions.PPELPE.get_items_ppelpe(items_popularity,items_entropy)
        self.items_bias = LogPopEnt.get_items_logpopent(
            items_popularity, items_entropy)
        print(self.items_bias.min(), self.items_bias.max())
        assert (self.items_bias.min() >= 0
                and np.isclose(self.items_bias.max(), 1))

        # regression_model = sklearn.linear_model.LinearRegression()
        res = scipy.optimize.minimize(
            lambda x, items_weights, items_bias: np.sum(
                (items_bias - x @ items_weights.T)**2),
            np.ones(self.num_latent_factors),
            args=(self.items_weights, self.items_bias),
            method='BFGS',
        )
        self.initial_b = res.x

        # print(np.corrcoef(self.items_bias,self.initial_b @ self.items_weights.T)[0,1])

        self.I = np.eye(len(self.items_weights[0]))
        self.bs = defaultdict(lambda: self.initial_b.copy())
        self.As = defaultdict(lambda: self.I.copy())
        items_score = _prediction_rule(self.I, self.initial_b,
                                       self.items_weights, self.alpha)

        # print("WSCB items score correlation with popularity:",scipy.stats.pearsonr(items_score,items_popularity),self.train_dataset.num_total_users, self.train_dataset.num_total_items)
        # print("WSCB items score correlation with entropy:",scipy.stats.pearsonr(items_score,items_entropy),self.train_dataset.num_total_users, self.train_dataset.num_total_items)
        # self.initial_b

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]
        user_model_items_score,l1,l2 = _prediction_rule(
            A, b, self.items_weights[candidate_items], self.alpha)
        items_score = user_model_items_score
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_latent_factors = self.items_weights[item]
        b = self.bs[uid]
        A = self.As[uid]
        A += max_item_latent_factors[:, None].dot(
            max_item_latent_factors[None, :])
        b += reward * max_item_latent_factors



"""
Cluster-Based algorithm of "Cluster-Based Bandits: Fast Cold-Start for Recommender System New Users"
"""
import numpy as np
from .ExperimentalValueFunction import ExperimentalValueFunction
import scipy.stats
from collections import defaultdict
from sklearn.cluster import KMeans
import itertools
import mf

# from joblib import Memory
from cachetools import cached
from cachetools.keys import hashkey


def _vars(a, axis=None):
    a_squared = a.copy()
    a_squared.data **= 2
    return a_squared.mean(axis) - np.square(a.mean(axis))


def _stds(a, axis=None):
    # print(_vars(a, axis))
    return np.sqrt(_vars(a, axis))


def _argmin(d):
    if not d:
        return None
    min_val = min(d.values())
    return [k for k in d if d[k] == min_val][0]


class CB(ExperimentalValueFunction):
    def __init__(
        self,
        num_clusters: int,
        B: float,
        C: float,
        D: float,
        num_lat: int,
        # cache_dir: str,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.B = B
        self.C = C
        self.D = D
        self.num_lat = num_lat
        self.num_clusters = num_clusters

        # self.cache_dir = cache_dir

        # self.Gamma = cached(self.Gamma,)

    # def find_object(db_handle, query):
    # print("processing {0}".format(query))
    # return query
    # memory = Memory(self.cache_dir, verbose=0)
    # self.Gamma =  memory.cache(self.Gamma, ignore=['self'])
    # # self.Rn =  memory.cache(self.Rn, ignore=['self'])
    # self.Sigma =  memory.cache(self.Sigma, ignore=['self'])
    # self.alpha =  memory.cache(self.alpha, ignore=['self'])
    @cached(cache={}, key=lambda self, g, h, v: hashkey(g, h, v))
    def Gamma(self, g, h, v):

        return (self.groups_mean[g][v] - self.groups_mean[h][v]) ** 2 / self.groups_std[
            g
        ][v]

    @cached(cache={}, key=lambda self, i, g, h: hashkey(i, g, h))
    def alpha(self, i, g, h):
        return self.Gamma(g, h, i) / np.sum(
            [self.Gamma(g, h, v) for v in range(self.num_total_items)]
        )

    @cached(cache={}, key=lambda self, g, h: hashkey(g, h))
    def Sigma(self, g, h):
        c = 0
        for v in range(self.num_total_items):
            c += self.Gamma(g, h, v)
        return c

    def explore(self, g, user_candidate_items):
        h = np.argmin([self.Sigma(g, h) for h in self.groups])
        return [self.Gamma(g, h, v) for v in user_candidate_items]

    def Rn(self, uid, g, h):
        # items = [
        # vi
        # for vi in range(self.num_total_items)
        # if self.groups_mean[g, vi] != self.groups_mean[h, vi]
        # ]
        items = np.arange(self.num_total_items)
        items = items[self.groups_mean[g, items] != self.groups_mean[h, items]]
        # self.[]
        alphas = np.array([self.alpha(vi, g, h) for vi in items])
        with np.errstate(invalid="ignore"):
            l = alphas * (
                (
                    self.consumption_matrix[uid, items].toarray().flatten()
                    - self.groups_mean[h, items].flatten()
                )
                / (
                    self.groups_mean[g, items].flatten()
                    - self.groups_mean[h, items].flatten()
                )
            )
        # print(np.sum(l))
        # l = [
        # self.alpha(vi, g, h)
        # * (
        # (self.consumption_matrix[uid, vi] - self.groups_mean[h, vi])
        # / (self.groups_mean[g, vi] - self.groups_mean[h, vi])
        # )
        # for vi in range(self.num_total_items)
        # if self.groups_mean[g, vi] != self.groups_mean[h, vi]
        # ]
        return np.sum(l)

    def I(self, uid, g):
        l = []
        for h in range(self.num_clusters):
            if g != h:
                l.append(self.Rn(uid, g, h))
        return np.min(l)

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset

        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (
                self.train_dataset.data[:, 2],
                (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1]),
            ),
            (self.train_dataset.num_total_users, self.train_dataset.num_total_items),
        )

        # self.train_consumption_matrix

        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        kmeans = KMeans(self.num_clusters).fit(mf_model.users_weights)
        self.num_total_items = self.train_dataset.num_total_items
        self.num_total_users = self.train_dataset.num_total_users

        self.groups_users = defaultdict(list)
        for uid, group in enumerate(kmeans.labels_):
            self.groups_users[group].append(uid)

        self.groups = list(range(self.num_clusters))
        self.groups_mean = np.zeros((self.num_clusters, self.num_total_items))
        self.groups_std = np.zeros((self.num_clusters, self.num_total_items))
        for group, uids in self.groups_users.items():
            ratings = self.train_consumption_matrix[uids]
            self.groups_mean[group] = ratings.mean(axis=0)
            num = np.array((ratings > 0).sum(axis=0)).flatten()
            # print(num.shape)
            try:
                self.groups_std[group] = _stds(ratings, axis=0)
            except:
                self.groups_std[group] = np.zeros(self.num_total_items)
            self.groups_std[group] += 0.5 * np.sqrt(np.log(1 / 0.2) / (num + 0.01))
            # print(group, self.groups_std[group])
        # self.groups_std[self.groups_std] +=

        self.consumption_matrix = self.train_consumption_matrix.todok()
        self.exploration_phase = defaultdict(lambda: True)
        self.new_user = defaultdict(lambda: True)
        self.users_group = {}

        # self.Gamma_cache = dict()
        # for g in range(self.num_clusters):
        # for h in range(self.num_clusters):
        # for v in range(self.num_total_items):
        # self.Gamma_cache[(g,h,v)] = self.Gamma(g,h,v)

        # self.Rn_cache = dict()
        # for g in range(self.num_clusters):
        # for h in range(self.num_clusters):
        # for u in range(self.num_total_users):
        # self.Rn_cache[(u,g,h)] = self.Rn(u,g,h)

        # print(self.groups)

        # for uid in range(self.train_dataset.data.shape[0]):
        # uid = int(self.train_dataset.data[uid, 0])
        # item = int(self.train_dataset.data[uid, 1])
        # reward = self.train_dataset.data[uid, 2]
        # # self.update(uid,item,reward,None)
        # self.update(None, (uid, item), reward, None)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        if self.new_user[uid]:
            g = np.random.randint(self.num_clusters)
            return self.explore(g, candidate_items), None
        else:
            # print("Not new user")
            if self.exploration_phase[uid]:
                # [g for g in self.groups if self.Rn(uid,g,)]
                candidates_groups = []
                # print("Not new user c1")
                for g in self.groups:
                    ngroups = set(self.groups) - {g}
                    hs_vals = []
                    for h in ngroups:
                        hs_vals.append(self.Rn(uid, g, h))
                    hs_vals = np.abs(hs_vals)
                    min_h = np.min(hs_vals)
                    fval = np.abs(min_h - 1)
                    if fval <= self.C:
                        candidates_groups.append(g)
                # print("Not new user c2")
                if len(candidates_groups) != 0:
                    g_hat = np.argmax([self.I(uid, g) for g in self.groups])
                    result_explore = self.explore(g_hat, candidate_items)
                    cond1 = (
                        np.min([self.Sigma(g_hat, h) for h in range(self.num_clusters)])
                        >= self.B
                    )
                    if cond1 or (
                        len(candidates_groups) == 1
                        and self.num_total_items > self.D * np.log2(self.num_clusters)
                    ):
                        self.users_group[uid] = g_hat
                        self.exploration_phase[uid] = False
                    return result_explore, None
                else:
                    sigma_values = dict()
                    for g, h in itertools.permutations(self.groups, 2):
                        sigma_values[(g, h)] = self.Sigma(g, h)
                    g, h = _argmin(sigma_values)
                    return self.explore(g, candidate_items), None
            else:
                user_g = self.users_group[uid]

                return [self.groups_mean[user_g][i] for i in candidate_items], None

        # return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        self.consumption_matrix[uid, item] = reward
        self.new_user[uid] = False
        # self.items_count[item] += 1



import numpy as np
import scipy
import scipy.stats
import sys, os
import random
from threadpoolctl import threadpool_limits
import ctypes
import metrics
from tqdm import tqdm

from irec.utils.utils import run_parallel
from mf import MF


class PMF(MF):
    def __init__(self,
                 iterations=200,
                 var=0.1,
                 user_var=1,
                 item_var=1,
                 learning_rate=1e-3,
                 momentum=0.6,
                 stop_criteria=0.0009,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.iterations = iterations
        self.var = var
        self.user_var = user_var
        self.item_var = item_var
        self.objective_values = []
        self.best = None
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.stop_criteria = stop_criteria

    def get_user_lambda(self):
        return self.var / self.user_var

    def get_item_lambda(self):
        return self.var / self.item_var

    def load_var(self, training_matrix):
        decimals = 4
        # if isinstance(training_matrix,scipy.sparse.spmatrix):
        #     non_empty_rows = np.sum(training_matrix>0,axis=1).A.flatten()
        #     training_matrix = training_matrix[non_empty_rows].A
        # training_matrix = self.normalize_matrix(training_matrix)

        # if not isinstance(training_matrix,scipy.sparse.spmatrix):
        #     self.var = np.var(training_matrix)
        #     self.user_var = np.mean(np.var(training_matrix,axis=1))
        #     self.item_var = np.mean(np.var(training_matrix,axis=0))
        # else:
        #     self.var = np.mean(training_matrix.data**2) - np.mean(training_matrix.data)**2
        #     self.user_var = np.mean([np.mean(i.data**2) - np.mean(i.data)**2 if i.getnnz()>0 else 0 for i in training_matrix])
        #     self.item_var = np.mean([np.mean(i.data**2) - np.mean(i.data)**2 if i.getnnz()>0 else 0 for i in training_matrix.transpose()])

        # # self.user_lambda = self.var/self.user_var
        # # self.item_lambda = self.var/self.item_var
        # self.var = np.round(self.var,decimals)
        # self.user_var = np.round(self.user_var,decimals)
        # self.item_var = np.round(self.item_var,decimals)
        # self.user_lambda = np.round(self.user_lambda,decimals)
        # self.item_lambda = np.round(self.item_lambda,decimals)

    def fit(self, training_matrix):
        super().fit()
        # training_matrix = self.normalize_matrix(training_matrix)
        decimals = 4
        user_lambda = self.var / self.user_var
        item_lambda = self.var / self.item_var
        num_users = training_matrix.shape[0]
        num_items = training_matrix.shape[1]
        lowest_value = np.min(training_matrix)
        highest_value = np.max(training_matrix)
        if not isinstance(training_matrix, scipy.sparse.spmatrix):
            indicator_matrix = training_matrix > lowest_value
            observed_ui = np.nonzero(indicator_matrix)
        else:
            observed_ui = (training_matrix.tocoo().row,
                           training_matrix.tocoo().col)
            # observed_ui = list(zip(*observed_ui))
        I = np.eye(self.num_lat)
        self.users_weights = 0.1 * np.random.rand(num_users, self.num_lat)
        self.items_weights = 0.1 * np.random.rand(num_items, self.num_lat)
        users_momentum = np.zeros(self.users_weights.shape)
        items_momentum = np.zeros(self.items_weights.shape)
        last_objective_value = np.NAN
        last_users_weights = self.users_weights
        last_items_weights = self.items_weights
        np.seterr('warn')
        predicted = self.predict(observed_ui)

        tq = tqdm(range(self.iterations))
        for i in tq:
            # print(f'[{i+1}/{self.iterations}]')
            error = scipy.sparse.csr_matrix(
                (training_matrix.data - predicted, observed_ui),
                shape=training_matrix.shape)
            users_gradient = error @ (
                -self.items_weights) + user_lambda * self.users_weights
            items_gradient = error.T @ (
                -self.users_weights) + item_lambda * self.items_weights

            users_momentum = self.momentum * users_momentum + self.learning_rate * users_gradient
            items_momentum = self.momentum * items_momentum + self.learning_rate * items_gradient

            self.users_weights -= users_momentum
            self.items_weights -= items_momentum

            predicted = self.predict(observed_ui)

            # objective_value = np.sum((training_matrix.data - predicted)**2)/2 +\
            #     user_lambda/2 * np.sum(np.linalg.norm(self.users_weights,axis=1)**2) +\
            #     item_lambda/2 * np.sum(np.linalg.norm(self.items_weights,axis=1)**2)
            # print("Objective value",objective_value)

            rmse = metrics.rmse(training_matrix.data, predicted)
            objective_value = rmse
            # print("RMSE",rmse)
            if objective_value > last_objective_value or np.fabs(
                    objective_value -
                    last_objective_value) <= self.stop_criteria:
                print(
                    "Achieved convergence with %d iterations, saving %d iteration"
                    % (i + 1, i))
                self.users_weights = last_users_weights
                self.items_weights = last_items_weights
                break

            last_objective_value = objective_value
            self.objective_value = objective_value
            last_users_weights = self.users_weights.copy()
            last_items_weights = self.items_weights.copy()

            tq.set_description('cur={:.3f},last={:.3f}'.format(
                objective_value, last_objective_value))



from threadpoolctl import threadpool_limits
from . import Recommender
import numpy as np
import mf
from collections import defaultdict
import util
import ctypes


class UserKNN(Recommender):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        pass

    def train(self, train_matrix):
        super().train()
        self.train_matrix = train_matrix
        self.num_neighbors = int(np.sqrt(train_matrix.shape[0]))
        np.seterr('warn')
        self.sim_matrix = np.corrcoef(train_matrix.A)
        np.seterr('raise')
        self.sim_matrix[np.isnan(self.sim_matrix)] = 0
        self.sim_matrix[self.sim_matrix < 0] = 0
        self.sim_matrix[self.sim_matrix == 0] = 0.001

        self.users_similar_users = dict()
        for uid in range(train_matrix.shape[0]):
            top_uids = np.argsort(self.sim_matrix[uid])[::-1]
            self.users_similar_users[uid] = top_uids[
                top_uids != uid][:self.num_neighbors]

        self.users_ratings_mean = np.zeros(train_matrix.shape[0])

        for uid in range(train_matrix.shape[0]):
            self.users_ratings_mean[uid] = np.mean(train_matrix[uid, :].data)

    def predict(self, test_matrix):
        super().predict()
        test_uids = np.nonzero(np.sum(test_matrix > 0, axis=1).A.flatten())[0]
        self_id = id(self)

        # print(len(test_matrix.data))
        # print(test_uids)

        with threadpool_limits(limits=1, user_api='blas'):
            args = [(
                self_id,
                int(uid),
            ) for uid in test_uids]
            results = util.run_parallel(self.predict_user_items, args)

        for i, user_result in enumerate(results):
            self.results[test_uids[i]] = user_result

        self.save_results()

    @staticmethod
    def predict_user_items(obj_id, uid):
        self = ctypes.cast(obj_id, ctypes.py_object).value
        train_matrix = self.train_matrix
        user_consumption = train_matrix[uid, :].A.flatten()
        consumed_iids = np.nonzero(user_consumption)[0]
        candidate_iids = np.where(user_consumption == 0)[0]

        candidate_items_score = np.zeros(candidate_iids.shape)
        neighboring_users = self.users_similar_users[uid]
        neighbors_similarities = self.sim_matrix[uid, neighboring_users]
        for i, candidate_iid in enumerate(candidate_iids):

            # consumed_neighbors = np.isin(neighboring_items,consumed_iids)
            # if np.count_nonzero(consumed_neighbors) == 0:
            #     candidate_items_score[i] = self.items_ratings_mean[candidate_iid]
            #     continue
            # neighbors_similarities[neighbors_similarities==0] = 0.001
            numerator = np.sum(neighbors_similarities *
                               (self.train_matrix[neighboring_users,
                                                  candidate_iid].A.flatten() -
                                self.users_ratings_mean[neighboring_users]))
            denominator = np.sum(neighbors_similarities)
            # try:
            candidate_items_score[
                i] = numerator / denominator + self.users_ratings_mean[uid]
            # except:
            #     print('numerator',numerator)
            #     print('denominator',denominator)
            #     print('ratings_mean',self.items_ratings_mean[candidate_iid])

        top_iids = candidate_iids[np.argsort(candidate_items_score)[::-1]]
        return top_iids[:self.result_list_size]

    def filter_parameters(self, parameters):
        return super().filter_parameters({
            k: v
            for k, v in parameters.items() if k not in ['num_neighbors']
        })



import itertools
from tqdm import tqdm
from util import Saveable
from sklearn.model_selection import KFold, StratifiedKFold
from collections import defaultdict
import scipy.sparse
import numpy as np


class GridSearch(Saveable):
    def __init__(self, model, parameters, n_splits=3):
        self.model = model
        self.parameters = parameters
        self.n_splits = n_splits

    def fit(self, X):
        combinations = list(itertools.product(*list(self.parameters.values())))
        parameters_names = list(self.parameters.keys())
        num_combinations = len(combinations)
        objective_values = defaultdict(list)

        if self.n_splits:
            kf = StratifiedKFold(n_splits=self.n_splits)
            X_data = X.data
            X_row = X.tocoo().row
            X_col = X.tocoo().col

        for i, values in enumerate(combinations):
            print(f'[{i+1}/{num_combinations}]')
            for parameter_name, value in zip(parameters_names, values):
                setattr(self.model, parameter_name, value)
            # print(list(zip(parameters_names,values)))

            if self.n_splits:
                for train_index, test_index in kf.split(X_data, y=X_row):
                    X_train = scipy.sparse.csr_matrix(
                        (X_data[train_index],
                         (X_row[train_index], X_col[train_index])),
                        shape=X.shape)
                    X_test = scipy.sparse.csr_matrix(
                        (X_data[test_index],
                         (X_row[test_index], X_col[test_index])),
                        shape=X.shape)
                    self.model.fit(X_train)
                    score = self.model.score(X_test)
                    objective_values[values].append(score)
                    print('Score:', score)
            else:
                self.model.fit(X)
                objective_values[values].append(self.model.score(X))

        print(objective_values)
        print(
            sorted({k: np.sum(v)
                    for k, v in objective_values.items()}.items(),
                   key=lambda x: x[1],
                   reverse=True))



"""
Code mainly from: https://github.com/zoulixin93/NICF
Author: zoulixin93
"""
from threadpoolctl import threadpool_limits
import numpy as np
from tqdm import tqdm
from .ExperimentalValueFunction import ExperimentalValueFunction
import matplotlib.pyplot as plt
import os
import scipy.sparse
from collections import defaultdict
import random
from .MFValueFunction import MFValueFunction
from numba import njit, jit
import mf
import joblib
import tensorflow as tf
import copy
from collections import defaultdict

MEMORYSIZE = 50000
BATCHSIZE = 128
THRESHOLD = 300

start = 0
end = 3000


def decay_function1(x):
    x = 50 + x
    return max(2.0 / (1 + np.power(x, 0.2)), 0.001)


START = decay_function1(start)
END = decay_function1(end)


def decay_function(x):
    x = max(min(end, x), start)
    return (decay_function1(x) - END) / (START - END + 0.0000001)


class Namespace:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class basic_model(object):
    GRAPHS = {}
    SESS = {}
    SAVER = {}

    def c_opt(self, learning_rate, name):
        if str(name).__contains__("adam"):
            print("adam")
            optimizer = tf.train.AdamOptimizer(learning_rate)
        elif str(name).__contains__("adagrad"):
            print("adagrad")
            optimizer = tf.train.AdagradOptimizer(learning_rate)
        elif str(name).__contains__("sgd"):
            optimizer = tf.train.GradientDescentOptimizer(learning_rate)
        elif str(name).__contains__("rms"):
            optimizer = tf.train.RMSPropOptimizer(learning_rate)
        elif str(name).__contains__("moment"):
            optimizer = tf.train.MomentumOptimizer(learning_rate, momentum=0.1)
        return optimizer

    @classmethod
    def create_model(
        cls,
        config,
        variable_scope="target",
        trainable=True,
        graph_name="DEFAULT",
        task_index=0,
    ):
        jobs = config.jobs
        job = list(jobs.keys())[0]
        cls.CLUSTER = tf.train.ClusterSpec(jobs)
        cls.SERVER = tf.train.Server(
            cls.CLUSTER,
            job_name=job,
            task_index=task_index,
            config=tf.compat.v1.ConfigProto(
                gpu_options=tf.compat.v1.GPUOptions(allow_growth=True)
            ),
        )
        if not graph_name in cls.GRAPHS:
            cls.GRAPHS[graph_name] = tf.Graph()
        with cls.GRAPHS[graph_name].as_default():
            model = cls(config, variable_scope=variable_scope, trainable=trainable)
            if not graph_name in cls.SESS:
                cls.SESS[graph_name] = tf.compat.v1.Session(cls.SERVER.target)
                cls.SAVER[graph_name] = tf.compat.v1.train.Saver(max_to_keep=50)
            cls.SESS[graph_name].run(model.init)
        return {
            "graph": cls.GRAPHS[graph_name],
            "sess": cls.SESS[graph_name],
            "saver": cls.SAVER[graph_name],
            "model": model,
            "cluster": cls.CLUSTER,
            "server": cls.SERVER,
        }

    @classmethod
    def create_model_without_distributed(
        cls, config, variable_scope="target", trainable=True, graph_name="DEFAULT"
    ):
        # if not graph_name in cls.GRAPHS:
        cls.GRAPHS[graph_name] = tf.Graph()
        with cls.GRAPHS[graph_name].as_default():
            model = cls(config, variable_scope=variable_scope, trainable=trainable)
            # if not graph_name in cls.SESS:
            cls.SESS[graph_name] = tf.compat.v1.Session(
                config=tf.compat.v1.ConfigProto(
                    gpu_options=tf.compat.v1.GPUOptions(allow_growth=True)
                )
            )
            cls.SAVER[graph_name] = tf.compat.v1.train.Saver(max_to_keep=50)
            cls.SESS[graph_name].run(model.init)
        return {
            "graph": cls.GRAPHS[graph_name],
            "sess": cls.SESS[graph_name],
            "saver": cls.SAVER[graph_name],
            "model": model,
        }

    def _update_placehoders(self):
        self.placeholders = {"none": {}}
        raise NotImplemented

    def _get_feed_dict(self, task, data_dicts):
        place_holders = self.placeholders[task]
        res = {}
        for key, value in place_holders.items():
            res[value] = data_dicts[key]
        return res

    def __init__(self, args, variable_scope="target", trainable=True):
        print(self.__class__)
        self.args = args
        self.variable_scope = variable_scope
        self.trainable = trainable
        self.placeholders = {}
        self._build_model()

    def _build_model(self):
        with tf.compat.v1.variable_scope(self.variable_scope):
            self._create_placeholders()
            self._create_global_step()
            self._update_placehoders()
            self._create_inference()
            if self.trainable:
                self._create_optimizer()
            self._create_intializer()

    def _create_global_step(self):
        self.global_step = tf.Variable(
            0, dtype=tf.int32, trainable=False, name="global_step"
        )

    def _create_intializer(self):
        with tf.name_scope("initlializer"):
            self.init = tf.compat.v1.global_variables_initializer()

    def _create_placeholders(self):
        raise NotImplementedError

    def _create_inference(self):
        raise NotImplementedError

    def _create_optimizer(self):
        raise NotImplementedError

    def chose_action(self, state, sess):
        raise NotImplementedError
        pass


def normalize(inputs, epsilon=1e-8, scope="ln", reuse=None):
    """Applies layer normalization.

    Args:
      inputs: A tensor with 2 or more dimensions, where the first dimension has
        `batch_size`.
      epsilon: A floating number. A very small number for preventing ZeroDivision Error.
      scope: Optional scope for `variable_scope`.
      reuse: Boolean, whether to reuse the weights of a previous layer
        by the same name.

    Returns:
      A tensor with the same shape and data dtype as `inputs`.
    """
    with tf.compat.v1.variable_scope(scope, reuse=reuse):
        inputs_shape = inputs.get_shape()
        params_shape = inputs_shape[-1:]

        mean, variance = tf.compat.v1.nn.moments(inputs, [-1], keep_dims=True)
        beta = tf.Variable(tf.zeros(params_shape))
        gamma = tf.Variable(tf.ones(params_shape))
        normalized = (inputs - mean) / ((variance + epsilon) ** (0.5))
        outputs = gamma * normalized + beta

    return outputs


def embedding(
    inputs,
    vocab_size,
    num_units,
    zero_pad=True,
    scale=True,
    l2_reg=0.0,
    scope="embedding",
    with_t=False,
    reuse=None,
):
    """Embeds a given tensor.

    Args:
      inputs: A `Tensor` with type `int32` or `int64` containing the ids
         to be looked up in `lookup table`.
      vocab_size: An int. Vocabulary size.
      num_units: An int. Number of embedding hidden units.
      zero_pad: A boolean. If True, all the values of the fist row (id 0)
        should be constant zeros.
      scale: A boolean. If True. the outputs is multiplied by sqrt num_units.
      scope: Optional scope for `variable_scope`.
      reuse: Boolean, whether to reuse the weights of a previous layer
        by the same name.

    Returns:
      A `Tensor` with one more rank than inputs's. The last dimensionality
        should be `num_units`.

    For example,

    ```
    import tensorflow as tf

    inputs = tf.to_int32(tf.reshape(tf.range(2*3), (2, 3)))
    outputs = embedding(inputs, 6, 2, zero_pad=True)
    with tf.compat.v1.Session() as sess:
        sess.run(tf.global_variables_initializer())
        print sess.run(outputs)
    >>
    [[[ 0.          0.        ]
      [ 0.09754146  0.67385566]
      [ 0.37864095 -0.35689294]]

     [[-1.01329422 -1.09939694]
      [ 0.7521342   0.38203377]
      [-0.04973143 -0.06210355]]]
    ```

    ```
    import tensorflow as tf

    inputs = tf.to_int32(tf.reshape(tf.range(2*3), (2, 3)))
    outputs = embedding(inputs, 6, 2, zero_pad=False)
    with tf.compat.v1.Session() as sess:
        sess.run(tf.global_variables_initializer())
        print sess.run(outputs)
    >>
    [[[-0.19172323 -0.39159766]
      [-0.43212751 -0.66207761]
      [ 1.03452027 -0.26704335]]

     [[-0.11634696 -0.35983452]
      [ 0.50208133  0.53509563]
      [ 1.22204471 -0.96587461]]]
    ```
    """
    with tf.variable_scope(scope, reuse=reuse):
        lookup_table = tf.get_variable(
            "lookup_table",
            dtype=tf.float32,
            shape=[vocab_size, num_units],
            # initializer=tf.contrib.layers.xavier_initializer(),
            regularizer=tf.contrib.layers.l2_regularizer(l2_reg),
        )
        if zero_pad:
            lookup_table = tf.concat(
                (tf.zeros(shape=[1, num_units]), lookup_table[1:, :]), 0
            )
        outputs = tf.nn.embedding_lookup(lookup_table, inputs)

        if scale:
            outputs = outputs * (num_units ** 0.5)
    if with_t:
        return outputs, lookup_table
    else:
        return outputs


def multihead_attention(
    queries,
    keys,
    num_units=None,
    num_heads=8,
    dropout_rate=0,
    is_training=True,
    causality=False,
    scope="multihead_attention",
    reuse=None,
    with_qk=False,
):
    """Applies multihead attention.

    Args:
      queries: A 3d tensor with shape of [N, T_q, C_q].
      keys: A 3d tensor with shape of [N, T_k, C_k].
      num_units: A scalar. Attention size.
      dropout_rate: A floating point number.
      is_training: Boolean. Controller of mechanism for dropout.
      causality: Boolean. If true, units that reference the future are masked.
      num_heads: An int. Number of heads.
      scope: Optional scope for `variable_scope`.
      reuse: Boolean, whether to reuse the weights of a previous layer
        by the same name.

    Returns
      A 3d tensor with shape of (N, T_q, C)
    """
    with tf.compat.v1.variable_scope(scope, reuse=reuse):
        # Set the fall back option for num_units
        if num_units is None:
            num_units = queries.get_shape().as_list[-1]

        # Linear projections
        # Q = tf.layers.dense(queries, num_units, activation=tf.nn.relu) # (N, T_q, C)
        # K = tf.layers.dense(keys, num_units, activation=tf.nn.relu) # (N, T_k, C)
        # V = tf.layers.dense(keys, num_units, activation=tf.nn.relu) # (N, T_k, C)
        Q = tf.compat.v1.layers.dense(
            queries, num_units, activation=None
        )  # (N, T_q, C)
        K = tf.compat.v1.layers.dense(keys, num_units, activation=None)  # (N, T_k, C)
        V = tf.compat.v1.layers.dense(keys, num_units, activation=None)  # (N, T_k, C)

        # Split and concat
        Q_ = tf.compat.v1.concat(
            tf.split(Q, num_heads, axis=2), axis=0
        )  # (h*N, T_q, C/h)
        K_ = tf.compat.v1.concat(
            tf.split(K, num_heads, axis=2), axis=0
        )  # (h*N, T_k, C/h)
        V_ = tf.compat.v1.concat(
            tf.split(V, num_heads, axis=2), axis=0
        )  # (h*N, T_k, C/h)

        # Multiplication
        outputs = tf.matmul(Q_, tf.transpose(K_, [0, 2, 1]))  # (h*N, T_q, T_k)

        # Scale
        outputs = outputs / (K_.get_shape().as_list()[-1] ** 0.5)

        # Key Masking
        key_masks = tf.sign(tf.abs(tf.reduce_sum(keys, axis=-1)))  # (N, T_k)
        key_masks = tf.tile(key_masks, [num_heads, 1])  # (h*N, T_k)
        key_masks = tf.tile(
            tf.expand_dims(key_masks, 1), [1, tf.shape(queries)[1], 1]
        )  # (h*N, T_q, T_k)

        paddings = tf.ones_like(outputs) * (-(2 ** 32) + 1)
        outputs = tf.where(tf.equal(key_masks, 0), paddings, outputs)  # (h*N, T_q, T_k)

        # Causality = Future blinding
        if causality:
            diag_vals = tf.ones_like(outputs[0, :, :])  # (T_q, T_k)
            tril = tf.linalg.LinearOperatorLowerTriangular(
                diag_vals
            ).to_dense()  # (T_q, T_k)
            masks = tf.tile(
                tf.expand_dims(tril, 0), [tf.shape(outputs)[0], 1, 1]
            )  # (h*N, T_q, T_k)

            paddings = tf.ones_like(masks) * (-(2 ** 32) + 1)
            outputs = tf.where(tf.equal(masks, 0), paddings, outputs)  # (h*N, T_q, T_k)

        # Activation
        outputs = tf.nn.softmax(outputs)  # (h*N, T_q, T_k)

        # Query Masking
        query_masks = tf.sign(tf.abs(tf.reduce_sum(queries, axis=-1)))  # (N, T_q)
        query_masks = tf.tile(query_masks, [num_heads, 1])  # (h*N, T_q)
        query_masks = tf.tile(
            tf.expand_dims(query_masks, -1), [1, 1, tf.shape(keys)[1]]
        )  # (h*N, T_q, T_k)
        outputs *= query_masks  # broadcasting. (N, T_q, C)

        # Dropouts
        outputs = tf.compat.v1.layers.dropout(
            outputs, rate=dropout_rate, training=tf.convert_to_tensor(is_training)
        )

        # Weighted sum
        outputs = tf.matmul(outputs, V_)  # ( h*N, T_q, C/h)

        # Restore shape
        outputs = tf.concat(tf.split(outputs, num_heads, axis=0), axis=2)  # (N, T_q, C)

        # Residual connection
        outputs += queries

        # Normalize
        # outputs = normalize(outputs) # (N, T_q, C)

    if with_qk:
        return Q, K
    else:
        return outputs


def feedforward(
    inputs,
    num_units=[2048, 512],
    scope="multihead_attention",
    dropout_rate=0.2,
    is_training=True,
    reuse=None,
):
    """Point-wise feed forward net.

    Args:
      inputs: A 3d tensor with shape of [N, T, C].
      num_units: A list of two integers.
      scope: Optional scope for `variable_scope`.
      reuse: Boolean, whether to reuse the weights of a previous layer
        by the same name.

    Returns:
      A 3d tensor with the same shape and dtype as inputs
    """
    with tf.compat.v1.variable_scope(scope, reuse=reuse):
        # Inner layer
        params = {
            "inputs": inputs,
            "filters": num_units[0],
            "kernel_size": 1,
            "activation": tf.nn.relu,
            "use_bias": True,
        }
        outputs = tf.compat.v1.layers.conv1d(**params)
        outputs = tf.compat.v1.layers.dropout(
            outputs, rate=dropout_rate, training=tf.convert_to_tensor(is_training)
        )
        # Readout layer
        params = {
            "inputs": outputs,
            "filters": num_units[1],
            "kernel_size": 1,
            "activation": None,
            "use_bias": True,
        }
        outputs = tf.compat.v1.layers.conv1d(**params)
        outputs = tf.compat.v1.layers.dropout(
            outputs, rate=dropout_rate, training=tf.convert_to_tensor(is_training)
        )

        # Residual connection
        outputs += inputs
    return outputs


class FA(basic_model):
    def _create_placeholders(self):
        self.utype = tf.compat.v1.placeholder(tf.int32, (None,), name="uid")
        self.p_rec = [
            tf.compat.v1.placeholder(
                tf.int32,
                (
                    None,
                    None,
                ),
                name="p" + str(i) + "_rec",
            )
            for i in range(6)
        ]
        self.pt = [
            tf.compat.v1.placeholder(tf.int32, (None, 2), "p" + str(i) + "t")
            for i in range(6)
        ]
        self.rec = tf.compat.v1.placeholder(tf.int32, (None,), name="iid")
        self.target = tf.compat.v1.placeholder(tf.float32, (None,), name="target")

    def _update_placehoders(self):
        self.placeholders["all"] = {
            "uid": self.utype,
            "iid": self.rec,
            "goal": self.target,
        }
        for i in range(6):
            self.placeholders["all"]["p" + str(i) + "_rec"] = self.p_rec[i]
            self.placeholders["all"]["p" + str(i) + "t"] = self.pt[i]
        self.placeholders["predict"] = {
            item: self.placeholders["all"][item]
            for item in ["uid"]
            + ["p" + str(i) + "_rec" for i in range(6)]
            + ["p" + str(i) + "t" for i in range(6)]
        }
        self.placeholders["optimize"] = self.placeholders["all"]

    def _create_inference(self):
        p_f = [
            tf.Variable(
                np.random.uniform(
                    -0.01, 0.01, (self.args.item_num, self.args.latent_factor)
                ),
                dtype=tf.float32,
                trainable=True,
                name="item" + str(i) + "_feature",
            )
            for i in range(6)
        ]
        u_f = tf.Variable(
            np.random.uniform(
                -0.01, 0.01, (self.args.utype_num, self.args.latent_factor)
            ),
            dtype=tf.float32,
            trainable=True,
            name="user_feature",
        )
        u_emb = tf.nn.embedding_lookup(u_f, self.utype)
        self.p_rec = [tf.transpose(item, [1, 0]) for item in self.p_rec]
        i_p_mask = [
            tf.expand_dims(tf.compat.v1.to_float(tf.not_equal(item, 0)), -1)
            for item in self.p_rec
        ]

        self.p_seq = [tf.nn.embedding_lookup(p_f[i], self.p_rec[i]) for i in range(6)]
        for iii, item in enumerate(self.p_seq):
            for i in range(self.args.num_blocks):
                with tf.compat.v1.variable_scope(
                    "rate_" + str(iii) + "_num_blocks_" + str(i)
                ):
                    item = multihead_attention(
                        queries=normalize(item),
                        keys=item,
                        num_units=self.args.latent_factor,
                        num_heads=self.args.num_heads,
                        dropout_rate=self.args.dropout_rate,
                        is_training=True,
                        causality=True,
                        scope="self_attention_pos_" + str(i),
                    )

                    item = feedforward(
                        normalize(item),
                        num_units=[self.args.latent_factor, self.args.latent_factor],
                        dropout_rate=self.args.dropout_rate,
                        is_training=True,
                        scope="feed_forward_pos_" + str(i),
                    )
                    item *= i_p_mask[iii]
        self.p_seq = [normalize(item) for item in self.p_seq]

        p_out = [
            tf.gather_nd(tf.transpose(self.p_seq[i], [1, 0, 2]), self.pt[i])
            for i in range(6)
        ]
        context = tf.concat(p_out, 1)
        hidden = tf.compat.v1.layers.dense(
            context, self.args.latent_factor, activation=tf.nn.relu
        )
        self.pi = tf.compat.v1.layers.dense(hidden, self.args.item_num, trainable=True)

    def _build_actor(self, context, name, trainable):
        with tf.compat.v1.variable_scope(name):
            a_prob = tf.layers.dense(context, self.args.item_num, trainable=trainable)
        params = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=name)
        return a_prob, params

    def _create_optimizer(self):
        a_indices = tf.stack(
            [tf.range(tf.shape(self.rec)[0], dtype=tf.int32), self.rec], axis=1
        )
        self.npi = tf.gather_nd(params=self.pi, indices=a_indices)
        self.loss = tf.losses.mean_squared_error(self.npi, self.target)
        self.optimizer = tf.compat.v1.train.AdamOptimizer(
            self.args.learning_rate
        ).minimize(self.loss)

    def optimize_model(self, sess, data):
        feed_dicts = self._get_feed_dict("optimize", data)
        return sess.run([self.loss, self.npi, self.optimizer], feed_dicts)[:2]

    def predict(self, sess, data):
        feed_dicts = self._get_feed_dict("predict", data)
        return sess.run(self.pi, feed_dicts)


def convert_item_seq2matrix(item_seq):
    max_length = max([len(item) for item in item_seq])
    matrix = np.zeros((max_length, len(item_seq)), dtype=np.int32)
    for x, xx in enumerate(item_seq):
        for y, yy in enumerate(xx):
            matrix[y, x] = yy
    target_index = list(zip([len(i) - 1 for i in item_seq], range(len(item_seq))))
    return matrix, target_index


class NICF(ExperimentalValueFunction):
    """NICF.
    
    It is an interactive method based on a combination of neural networks and 
    collaborative filtering that also performs a meta-learning of the user’s preferences [1]_.

    References
    ----------
    .. [1] Zhao, Xiaoxue, Weinan Zhang, and Jun Wang. "Interactive collaborative filtering." 
       Proceedings of the 22nd ACM international conference on Information & Knowledge Management. 2013.   
    """
    def __init__(
        self,
        time_step,
        latent_factor,
        learning_rate,
        training_epoch,
        rnn_layer,
        inner_epoch,
        batch,
        gamma,
        clip_param,
        restore_model,
        num_blocks,
        num_heads,
        dropout_rate,
        *args,
        **kwargs
    ):
        """__init__.

        Args:
            args:
            kwargs:
            time_step:
            latent_factor:
            learning_rate:
            training_epoch:
            rnn_layer:
            inner_epoch:
            batch:
            gamma:
            clip_param:
            restore_model:
            num_blocks:
            num_heads:
            dropout_rate:
        """

        super().__init__(*args, **kwargs)
        self.time_step = time_step
        self.latent_factor = latent_factor
        self.learning_rate = learning_rate
        self.training_epoch = training_epoch
        self.rnn_layer = rnn_layer
        self.inner_epoch = inner_epoch
        self.batch = batch
        self.gamma = gamma
        self.clip_param = clip_param
        self.restore_model = restore_model
        self.num_blocks = num_blocks
        self.num_heads = num_heads
        self.dropout_rate = dropout_rate

    def reset_with_users(self, uid):
        self.state = [(uid, 1), []]
        self.short = {}
        return self.state

    def step(self, action):
        if action in self.rates[self.state[0][0]] and (not action in self.short):
            rate = self.rates[self.state[0][0]][action]
            if rate >= 4:
                reward = 1
            else:
                reward = 0
        else:
            rate = 0
            reward = 0

        if len(self.state[1]) < self.time_step - 1:
            done = False
        else:
            done = True
        self.short[action] = 1
        t = self.state[1] + [[action, reward, done]]
        info = rate
        self.state[1].append([action, reward, done, info])
        return self.state, reward, done, info

    def train_epoch(self, epoch):
        selected_users = np.random.choice(self.train_dataset.uids, (self.inner_epoch,))
        for uuid in selected_users:
            actions = {}
            done = False
            state = self.reset_with_users(uuid)
            while not done:
                data = {"uid": [state[0][1]]}
                for i in range(6):
                    p_r, pnt = convert_item_seq2matrix(
                        [[0] + [item[0] for item in state[1] if item[3] == i]]
                    )
                    data["p" + str(i) + "_rec"] = p_r
                    data["p" + str(i) + "t"] = pnt
                policy = self.fa["model"].predict(self.fa["sess"], data)[0]
                if np.random.random() < 5 * THRESHOLD / (THRESHOLD + self.tau):
                    policy = np.random.uniform(0, 1, (self.args.item_num,))
                for item in actions:
                    policy[item] = -np.inf
                action = np.argmax(policy[1:]) + 1
                s_pre = copy.deepcopy(state)
                state_next, rwd, done, info = self.step(action)
                self.memory.append(
                    [s_pre, action, rwd, done, copy.deepcopy(state_next)]
                )
                actions[action] = 1
                state = state_next

        if len(self.memory) >= BATCHSIZE:
            self.memory = self.memory[-MEMORYSIZE:]
            batch = [
                self.memory[item]
                for item in np.random.choice(range(len(self.memory)), (BATCHSIZE,))
            ]
            data = self.convert_batch2dict(batch, epoch)
            loss, _ = self.fa["model"].optimize_model(self.fa["sess"], data)
            self.tau += 5

    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """ 
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = copy.copy(train_dataset)
        self.train_dataset.data[:, 2]
        self.train_dataset.data[:, 2] = np.ceil(self.train_dataset.data[:, 2])
        self.train_dataset.data[:, 0] += 1
        self.train_dataset.data[:, 1] += 1
        self.train_dataset.data = self.train_dataset.data.astype(int)
        self.train_dataset.update_from_data()
        # self.train_consumption_matrix = scipy.sparse.csr_matrix((self.train_dataset.data[:,2],(self.train_dataset.data[:,0],self.train_dataset.data[:,1])),(self.train_dataset.num_total_users,self.train_dataset.num_total_items))
        self.tau = 0

        args = Namespace(
            **{i: getattr(self, i) for i in dir(self)},
            item_num=self.train_dataset.num_total_items + 1,
            utype_num=self.train_dataset.num_total_users + 1
        )
        self.args = args
        self.fa = FA.create_model_without_distributed(args)
        self.memory = []
        self.rates = defaultdict(dict)
        for i in range(len(self.train_dataset.data)):
            uid = int(self.train_dataset.data[i, 0])
            item = int(self.train_dataset.data[i, 1])
            reward = self.train_dataset.data[i, 2]
            self.rates[uid][item] = reward

        self.rates = dict(self.rates)

        for epoch in tqdm(range(self.training_epoch)):
            self.train_epoch(epoch)

        self.test_users_states = dict()

    def _update(self, uid, item, reward):
        pass

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)
        
        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        uid += 1
        if uid not in self.test_users_states:
            self.test_users_states[uid] = [(uid, 1), []]
        state = self.test_users_states[uid]
        data = {"uid": [state[0][1]]}
        for i in range(6):
            p_r, pnt = convert_item_seq2matrix(
                [[0] + [item[0] for item in state[1] if item[3] == i]]
            )
            data["p" + str(i) + "_rec"] = p_r
            data["p" + str(i) + "t"] = pnt

        policy = self.fa["model"].predict(self.fa["sess"], data)[0]
        items_score = policy[1:][candidate_items]
        # for item in candidate_items:
        # policy[item] = -np.inf
        # items_score=np.random.rand(len(candidate_items))
        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        uid += 1
        item += 1
        if uid not in self.test_users_states:
            self.test_users_states[uid] = [(uid, 1), []]
        state = self.test_users_states[uid]
        self.state = state

        action = item
        done = False
        rate = reward
        if rate >= 4:
            reward = 1
        else:
            reward = 0

        t = self.state[1] + [[action, reward, done]]
        info = rate
        self.state[1].append([action, reward, done, info])

    def convert_batch2dict(self, batch, epoch):
        uids = []
        pos_recs = {i: [] for i in range(6)}
        next_pos = {i: [] for i in range(6)}
        iids = []
        goals = []
        dones = []
        for item in batch:
            uids.append(item[0][0][1])
            ep = item[0][1]
            for xxx in range(6):
                pos_recs[xxx].append([0] + [j[0] for j in ep if j[3] == xxx])
            iids.append(item[1])
            goals.append(item[2])
            if item[3]:
                dones.append(0.0)
            else:
                dones.append(1.0)
            ep = item[4][1]
            for xxx in range(6):
                next_pos[xxx].append([0] + [j[0] for j in ep if j[3] == xxx])
        data = {"uid": uids}
        for xxx in range(6):
            p_r, pnt = convert_item_seq2matrix(next_pos[xxx])
            data["p" + str(xxx) + "_rec"] = p_r
            data["p" + str(xxx) + "t"] = pnt
        value = self.fa["model"].predict(self.fa["sess"], data)
        value[:, 0] = -500
        goals = (
            np.max(value, axis=-1)
            * np.asarray(dones)
            * min(self.args.gamma, decay_function(max(end - epoch, 0) + 1))
            + goals
        )
        data = {"uid": uids, "iid": iids, "goal": goals}
        for i in range(6):
            p_r, pnt = convert_item_seq2matrix(pos_recs[i])
            data["p" + str(i) + "_rec"] = p_r
            data["p" + str(i) + "t"] = pnt
        return data

    def precision(self, episode):
        return sum([i[1] for i in episode])

    def recall(self, episode, uid):
        return sum([i[1] for i in episode]) / len(self.rates[uid])



from os.path import dirname, realpath, sep, pardir
import sys, os
sys.path.append(dirname(realpath(__file__)) + sep + pardir)

import numpy as np
import scipy.sparse

# from .. import irec.utils
from collections import defaultdict
import pickle
import json
from .ValueFunction import ValueFunction


class ExperimentalValueFunction(ValueFunction):
    def __init__(self, *args, **kwargs):
        ValueFunction.__init__(self, *args, **kwargs)

    def reset(self, observation):
        train_dataset = observation
        ValueFunction.reset(self, train_dataset)



from irec.RelevanceEvaluator import RelevanceEvaluator
import numpy as np

import irec.value_functions

import scipy.sparse
from collections import defaultdict
import time
from irec.utils.utils import run_parallel
import ctypes

# import irec.utils.dataset as dataset
from irec.utils import dataset
from irec.value_functions.Entropy import Entropy
from typing import Any

np.seterr(all="raise")

"""Evaluation Metrics.

This module implements numerous evaluation metrics widely used in RS.
"""

class Metric:
    """Metric.

    Metrics are used to assess the performance of a recommendation system. 
    For this, there are several metrics capable of evaluating recommendations in different ways.    
    """

    def __init__(
        self,
        ground_truth_dataset: Any,
        relevance_evaluator: RelevanceEvaluator,
        *args,
        **kwargs
    ):
        """__init__.

        Args:
            ground_truth_dataset (Any): ground_truth_dataset
            relevance_evaluator (RelevanceEvaluator): relevance_evaluator
            args:
            kwargs:
        """

        super().__init__(*args, **kwargs)
        self.ground_truth_dataset = ground_truth_dataset
        self.relevance_evaluator = relevance_evaluator

    def compute(self, uid: int) -> Any:
        """compute.
        
        This method performs the metric calculation for a given user.

        Args:
            uid (int): uid

        Returns:
            Any:
        """

        return None

    def update_recommendation(self, uid: int, item: int, reward: float) -> None:
        """update_recommendation.
        
        Uses user-supplied item rating to update metric attributes.

        Args:
            uid (int): uid
            item (int): item
            reward (float): reward

        Returns:
            None:
        """
        raise NotImplemented

    def update_consumption_history(self, uid: int, item: int, reward: float) -> None:
        """update_consumption_history.
        
        Update items consumed by a user.
        
        Args:
            uid (int): uid
            item (int): item
            reward (float): reward

        Returns:
            None:
        """

        raise NotImplemented


class Recall(Metric):
    """Recall.
    
    Recall represents the probability that a relevant item will be selected. 
    (true positive/false negative)
    """

    def __init__(self, users_false_negative, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            users_false_negative:
        """

        super().__init__(*args, **kwargs)
        self.users_true_positive = defaultdict(int)
        self.users_false_negative = users_false_negative

    def compute(self, uid: int):
        """compute.

        Args:
            uid (int): user id
        """
        if self.users_true_positive[uid] == 0 and self.users_false_negative[uid] == 0:
            return 0
        return self.users_true_positive[uid] / self.users_false_negative[uid]

    def update_recommendation(self, uid: int, item: int, reward: float):
        """update_recommendation.

        Args:
            uid (int): user id
            item (int): item id
            reward (float): reward
        """
        if self.relevance_evaluator.is_relevant(reward):
            self.users_true_positive[uid] += 1


class Precision(Metric):
    """Precision.

    Precision is defined as the percentage of predictions we get right.
    (true positive)/(total predictions).
    """

    def __init__(self, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)
        self.users_true_positive = defaultdict(int)
        self.users_false_positive = defaultdict(int)

    def compute(self, uid: int):
        """compute.

        Args:
            uid (int): user id
        """
        if self.users_true_positive[uid] == 0 and self.users_false_positive[uid] == 0:
            return 0
        return self.users_true_positive[uid] / (
            self.users_true_positive[uid] + self.users_false_positive[uid]
        )

    def update_recommendation(self, uid: int, item: int, reward: float):
        """update_recommendation.

        Args:
            uid (int): user id
            item (int): item id
            reward (float): reward
        """
        if self.relevance_evaluator.is_relevant(reward):
            self.users_true_positive[uid] += 1
        else:
            self.users_false_positive[uid] += 1


class Hits(Metric):
    """Hits.
    
    Number of recommendations made successfully.
    (right predictions)
    """

    def __init__(self, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)
        self.users_true_positive = defaultdict(int)

    def compute(self, uid: int):
        """compute.

        Args:
            uid (int): user id
        """
        return self.users_true_positive[uid]

    def update_recommendation(self, uid: int, item: int, reward: float):
        """update_recommendation.

        Args:
            uid (int): user id
            item (int): item id
            reward (float): reward
        """
        if self.relevance_evaluator.is_relevant(reward):
            self.users_true_positive[uid] += 1


class NumInteractions(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.users_num_interactions = defaultdict(int)

    def compute(self, uid):
        return self.users_num_interactions[uid]

    def update_recommendation(self, uid, item, reward):
        self.users_num_interactions[uid] += 1


class EPC(Metric):
    """Expected Popularity Complement.
    
    EPC is a metric that measures the ability of a system to recommend 
    relevant items that reside in the long-tail.
    """
   
    def __init__(self, items_normalized_popularity, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            items_normalized_popularity
        """
        super().__init__(*args, **kwargs)
        self.users_num_items_recommended = defaultdict(int)
        self.users_prob_not_seen_cumulated = defaultdict(float)
        self.items_normalized_popularity = items_normalized_popularity

    def compute(self, uid: int):
        """compute.

        Args:
            uid (int): user id
        """
        C_2 = 1.0 / self.users_num_items_recommended[uid]
        sum_2 = self.users_prob_not_seen_cumulated[uid]
        EPC = C_2 * sum_2
        return EPC

    def update_recommendation(self, uid: int, item: int, reward: float):
        """update_recommendation.

        Args:
            uid (int): user id
            item (int): item id
            reward (float): reward
        """        
        self.users_num_items_recommended[uid] += 1
        probability_seen = self.items_normalized_popularity[item]
        self.users_prob_not_seen_cumulated[uid] += 1 - probability_seen


class Entropy(Metric):
    def __init__(self, items_entropy, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.users_num_items_recommended = defaultdict(int)
        self.users_entropy_cumulated = defaultdict(float)
        self.items_entropy = items_entropy

    def compute(self, uid):
        return self.users_entropy_cumulated[uid] / self.users_num_items_recommended[uid]

    def update_recommendation(self, uid, item, reward):
        self.users_num_items_recommended[uid] += 1
        self.users_entropy_cumulated[uid] += self.items_entropy[item]


class TopItemsMembership(Metric):
    def __init__(self, items_feature_values, top_size, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.users_num_items_recommended = defaultdict(int)
        self.users_membership_count_cumulated = defaultdict(float)
        self.items_feature_values = items_feature_values
        self.top_size = top_size
        self.top_items = set(np.argsort(items_feature_values)[::-1][: self.top_size])

    def compute(self, uid):
        return (
            self.users_membership_count_cumulated[uid]
            / self.users_num_items_recommended[uid]
        )

    def update_recommendation(self, uid, item, reward):
        self.users_num_items_recommended[uid] += 1
        if item in self.top_items:
            self.users_membership_count_cumulated[uid] += 1


class ILD(Metric):
    """Intra-List Diversity.

    This is used to measure the diversity of an individual user’s recommendations and quantifies user-novelty.   
    """

    def __init__(self, items_distance, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            items_distance:
        """
        super().__init__(*args, **kwargs)
        self.items_distance = items_distance
        self.users_items_recommended = defaultdict(list)
        self.users_local_ild = defaultdict(float)

    def compute(self, uid: int):
        """compute.

        Args:
            uid (int): user id
        """
        user_num_items_recommended = len(self.users_items_recommended[uid])
        if user_num_items_recommended == 0 or user_num_items_recommended == 1:
            return 1.0
        else:
            return self.users_local_ild[uid] / (
                user_num_items_recommended * (user_num_items_recommended - 1) / 2
            )

    def update_recommendation(self, uid: int, item: int, reward: float):
        """update_recommendation.

        Args:
            uid (int): user id
            item (int): item id
            reward (float): reward
        """
        self.users_local_ild[uid] += np.sum(
            self.items_distance[self.users_items_recommended[uid], item]
        )
        self.users_items_recommended[uid].append(item)


class EPD:
    """Expected Profile Distance.

    EPD, on the other hand, is a distance-based novelty measure, which looks
    at distances between the items inthe user’s profile and the recommended items.   
    """
    def __init__(self, items_distance, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            items_distance:
        """

        super().__init__(*args, **kwargs)
        self.items_distance = items_distance
        self.users_consumed_items = defaultdict(list)
        self.users_items_recommended = defaultdict(list)
        self.users_local_ild = defaultdict(float)

        self.users_relevant_items = scipy.sparse.csr_matrix(self.ground_truth_dataset)
        self.users_relevant_items[
            self.users_relevant_items > self.ground_truth_dataset.min_rating
        ] = True

    def compute(self, uid: int):
        """compute.

        Args:
            uid (int): user id
        """
        rel = np.array(self.users_relevant_items[uid].A).flatten()
        consumed_items = self.users_consumed_items[item]
        predicted = self.users_items_recommended[uid]
        res = (
            rel[predicted][:, None]
            @ rel[consumed_items][None, :]
            * self.items_distance[predicted, :][:, consumed_items]
        )
        C = 1 / (len(predicted) * np.sum(rel[consumed_items]))
        return C * np.sum(res)

    def update_recommendation(self, uid: int, item: int, reward: float):
        """update_recommendation.

        Args:
            uid (int): user id
            item (int): item id
            reward (float): reward
        """
        self.users_local_ild[uid] += np.sum(
            self.items_distance[self.users_items_recommended[uid], item]
        )
        self.users_items_recommended[uid].append(item)

    def update_consumption_history(self, uid: int, item: int, reward: float):
        """update_consumption_history.

        Args:
            uid (int): user id
            item (int): item id
            reward (float): reward
        """
        self.users_consumed_items[uid].append(item)


class AP(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.users_true_positive = defaultdict(int)
        self.users_false_positive = defaultdict(int)
        self.users_cumulated_precision = defaultdict(float)
        self.users_num_recommendations = defaultdict(int)

    def compute(self, uid):
        return self.users_cumulated_precision[uid] / self.users_num_recommendations

    def update_recommendation(self, uid, item, reward):
        if self.relevance_evaluator.is_relevant(reward):
            self.users_true_positive[uid] += 1
        else:
            self.users_false_positive[uid] += 1

        self.users_cumulated_precision[uid] += self.users_true_positive[uid] / (
            self.users_true_positive[uid] + self.users_false_positive[uid]
        )
        self.users_num_recommendations[uid] += 1


class GiniCoefficientInv(Metric):
    """GiniCoefficientInv.
    
    desc    
    """

    def __init__(self, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            users_covered:
        """
        super().__init__(*args, **kwargs)
        self.items_frequency = defaultdict(int)
        for item in np.unique(self.ground_truth_dataset.data[:, 1]):
            self.items_frequency[item]
        self.is_computation_updated = False
        self.computation_cache = None

    def compute(self, uid: int):
        """compute.

        Args:
            uid (int): user id
        """
        if self.is_computation_updated == False:
            self.is_computation_updated = True
            x = np.array(list(self.items_frequency.values()))
            diff_sum = 0
            for i, xi in enumerate(x[:-1], 1):
                diff_sum += np.sum(np.abs(xi - x[i:]))
            self.computation_cache = diff_sum / (len(x) ** 2 * np.mean(x))
        return 1 - self.computation_cache

    def update_recommendation(self, uid: int, item: int, reward: float):
        """update_recommendation.

        Args:
            uid (int): user id
            item (int): item id
            reward (float): reward
        """
        self.items_frequency[item] += 1
        if self.is_computation_updated:
            self.is_computation_updated = False


class UsersCoverage(Metric):
    """Users Coverage.
    
    It represents the percentage of distinctusers that are interested 
    in at least k items recommended (k ≥ 1).
    """
    def __init__(self, users_covered=defaultdict(bool), *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            users_covered:
        """
        super().__init__(*args, **kwargs)
        self.users_covered = users_covered

    def compute(self, uid: int):
        """compute.

        Args:
            uid (int): user id
        """
        l = np.array(list(self.users_covered.values()))
        return np.sum(l) / len(l)

    def update_recommendation(self, uid: int, item: int, reward: float):
        """update_consumption_history.

        Args:
            uid (int): user id
            item (int): item id
            reward (float): reward
        """
        if self.users_covered[uid] == False and self.relevance_evaluator.is_relevant(
            reward
        ):
            self.users_covered[uid] = True
            # else:
            # self.users_covered[uid] = False


def mapk(actual, predicted, k):
    score = 0.0
    num_hits = 0.0

    for i, p in enumerate(predicted):
        if p in actual and p not in predicted[:i]:
            num_hits += 1.0
            score += num_hits / (i + 1.0)

    if not actual:
        return 0.0

    return score / min(len(actual), k)


def hitsk(actual, predicted):
    return len(set(predicted) & set(actual))


def precisionk(actual, predicted):
    return 1.0 * len(set(actual) & set(predicted)) / len(predicted)


def recallk(actual, predicted):
    return 1.0 * len(set(actual) & set(predicted)) / len(actual)


def f1k(precision, recall):
    if precision == 0 and recall == 0:
        return 0
    return 2 * (precision * recall) / (precision + recall)


def ndcgk(actual, predicted):
    idcg = 1.0
    dcg = 1.0 if predicted[0] in actual else 0.0
    for i, p in enumerate(predicted[1:]):
        if p in actual:
            dcg += 1.0 / np.log(i + 2)
        idcg += 1.0 / np.log(i + 2)
    return dcg / idcg


def epck(actual, predicted, items_popularity):
    C_2 = 1.0 / len(predicted)
    sum_2 = 0
    for i, lid in enumerate(predicted):
        # if lid in actual:
        prob_seen_k = items_popularity[lid]
        sum_2 += 1 - prob_seen_k
    EPC = C_2 * sum_2
    return EPC


def ildk(items, items_distance):
    items = np.array(items)
    num_items = len(items)
    local_ild = 0
    if num_items == 0 or num_items == 1:
        # print("Number of items:",num_items)
        return 1.0
    else:
        for i, item_1 in enumerate(items):
            for j, item_2 in enumerate(items):
                if j < i:
                    local_ild += items_distance[item_1, item_2]

    return local_ild / (num_items * (num_items - 1) / 2)


def get_items_distance(matrix):
    if isinstance(matrix, scipy.sparse.spmatrix):
        items_similarity = np.corrcoef(matrix.A.T)
        # matrix = matrix.T
        # center_matrix = matrix.sum(axis=1)
        # center_matrix = center_matrix @ center_matrix.T
        # cov_matrix = (matrix @ matrix.T - center_matrix)/(center_matrix.shape[0]-1)
        # cov_diag = np.diag(cov_matrix)
        # items_similarity = cov_matrix/np.sqrt(np.outer(cov_diag,cov_diag))
    else:
        items_similarity = np.corrcoef(matrix.T)
    items_similarity = (items_similarity + 1) / 2
    # items_similarity[items_similarity < 0] = 0
    return 1 - items_similarity


def epdk(actual, predicted, consumed_items, items_distance):
    if len(consumed_items) == 0:
        return 1
    rel = np.zeros(items_distance.shape[0], dtype=bool)
    rel[actual] = 1
    # distances_sum
    res = (
        rel[predicted][:, None]
        @ rel[consumed_items][None, :]
        * items_distance[predicted, :][:, consumed_items]
    )
    C = 1 / (len(predicted) * np.sum(rel[consumed_items]))
    return C * np.sum(res)


def rmse(ground_truth, predicted):
    return np.sqrt(np.mean((predicted - ground_truth) ** 2))



from threadpoolctl import threadpool_limits
import numpy as np
from tqdm import tqdm
from .ExperimentalValueFunction import ExperimentalValueFunction
import matplotlib.pyplot as plt
import os
import scipy.sparse
from collections import defaultdict
import random
from .MFValueFunction import MFValueFunction
from tqdm import tqdm
from numba import njit, jit
import mf
import joblib
import value_functions


@njit
def _softmax(x):
    return np.exp(x - np.max(x)) / np.sum(np.exp(x - np.max(x)))


class PTS(MFValueFunction):
    """Particle Thompson sampling.
    
    It is a PMF formulation for the original TS based on a Bayesian inference around the items.
    This method also applies particle filtering to guide the exploration of items over time [1]_.

    References
    ----------
    .. [1] Wang, Qing, et al. "Online interactive collaborative filtering using multi-armed 
       bandit with dependent arms." IEEE Transactions on Knowledge and Data Engineering 31.8 (2018): 1569-1580.   
    """
    
    def __init__(self, num_particles, var, var_u, var_v, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            num_particles (int):
            var:
            var_u:
            var_v:
        """
        super().__init__(*args, **kwargs)
        self.num_particles = num_particles
        self.var = var
        self.var_u = var_u
        self.var_v = var_v

    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """ 
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (
                self.train_dataset.data[:, 2],
                (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1]),
            ),
            (self.train_dataset.num_total_users, self.train_dataset.num_total_items),
        )

        self.num_total_items = self.train_dataset.num_total_items
        self.num_total_users = self.train_dataset.num_total_users

        self.particles_us = np.random.normal(
            size=(self.num_particles, self.num_total_users, self.num_lat)
        )
        self.particles_vs = np.random.normal(
            size=(self.num_particles, self.num_total_items, self.num_lat)
        )
        self.particles_var_us = np.ones(shape=(self.num_particles)) * self.var_u
        self.particles_var_vs = np.ones(shape=(self.num_particles)) * self.var_v

        self.particles_ids = np.arange(self.num_particles)
        self.items_consumed_users = defaultdict(list)
        self.items_consumed_users_rewards = defaultdict(list)
        self.users_consumed_items = defaultdict(list)
        self.users_consumed_items_rewards = defaultdict(list)
        for i in tqdm(range(len(self.train_dataset.data))):
            uid = int(self.train_dataset.data[i, 0])
            item = int(self.train_dataset.data[i, 1])
            reward = self.train_dataset.data[i, 2]
            # self.update(uid,item,reward,None)
            self.users_consumed_items[uid].append(item)
            self.users_consumed_items_rewards[uid].append(reward)
            self.items_consumed_users[item].append(uid)
            self.items_consumed_users_rewards[item].append(reward)

        mf_model = mf.PMF(
            num_lat=self.num_lat, var=self.var, user_var=self.var_u, item_var=self.var_v
        )
        # mf_model_id = joblib.hash(
        # (mf_model.get_id(), self.train_consumption_matrix))
        # pdm = PersistentDataManager('state_save')
        # if pdm.file_exists(mf_model_id):
        # mf_model = pdm.load(mf_model_id)
        # else:
        mf_model.fit(self.train_consumption_matrix)
        # pdm.save(mf_model_id, mf_model)

        for i in range(self.num_particles):
            self.particles_us[i] = mf_model.users_weights
            self.particles_vs[i] = mf_model.items_weights

        # for uid in self.users_consumed_items.keys():
        #     item = self.users_consumed_items[uid].pop()
        #     reward = self.users_consumed_items_rewards[uid].pop()
        #     self.update(uid,item,reward,None)

        # for item in self.items_consumed_users.keys():
        #     uid = self.items_consumed_users[item].pop()
        #     reward = self.items_consumed_users_rewards[item].pop()
        #     self.update(uid,item,reward,None)

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)
        
        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        particle_idx = np.random.choice(self.particles_ids)
        items_score = (
            self.particles_us[particle_idx][uid]
            @ self.particles_vs[particle_idx][candidate_items].T
        )
        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        with threadpool_limits(limits=1, user_api="blas"):
            updated_history = False

            lambdas_u_i = np.empty(
                shape=(self.num_particles, self.num_lat, self.num_lat)
            )
            zetas_u_i = np.empty(shape=(self.num_particles, self.num_lat))
            mus_u_i = np.empty(shape=(self.num_particles, self.num_lat))
            for i in range(self.num_particles):
                v_j = self.particles_vs[i][self.users_consumed_items[uid]]
                lambda_u_i = 1 / self.var * (v_j.T @ v_j) + 1 / self.particles_var_us[
                    i
                ] * np.eye(self.num_lat)
                zeta_u_i = np.sum(
                    np.multiply(
                        v_j,
                        np.array(self.users_consumed_items_rewards[uid]).reshape(-1, 1),
                    ),
                    axis=0,
                )
                lambdas_u_i[i] = lambda_u_i
                zetas_u_i[i] = zeta_u_i
                mus_u_i[i] = 1 / self.var * (np.linalg.inv(lambda_u_i) @ zeta_u_i)

            weights = np.empty(self.num_particles)
            for i in range(self.num_particles):
                lambda_u_i, mu_u_i = lambdas_u_i[i], mus_u_i[i]
                v_j = self.particles_vs[i][item, :]
                cov = 1 / self.var + np.dot(np.dot(v_j.T, lambda_u_i), v_j)
                w = scipy.stats.norm(np.dot(v_j.T, mu_u_i), cov).pdf(reward)
                weights[i] = w

            normalized_weights = _softmax(weights)
            ds = np.random.choice(
                range(self.num_particles), p=normalized_weights, size=self.num_particles
            )
            new_particles_us = np.empty(
                shape=(self.num_particles, self.num_total_users, self.num_lat)
            )
            new_particles_vs = np.empty(
                shape=(self.num_particles, self.num_total_items, self.num_lat)
            )
            new_particles_var_us = np.empty(shape=(self.num_particles))
            new_particles_var_vs = np.empty(shape=(self.num_particles))

            for i in range(self.num_particles):
                d = ds[i]
                new_particles_us[i] = self.particles_us[d]
                new_particles_vs[i] = self.particles_vs[d]
                new_particles_var_us[i] = self.particles_var_us[d]
                new_particles_var_vs[i] = self.particles_var_vs[d]

            if not updated_history:
                self.users_consumed_items[uid].append(item)
                self.users_consumed_items_rewards[uid].append(reward)
                self.items_consumed_users[item].append(uid)
                self.items_consumed_users_rewards[item].append(reward)
                updated_history = True

            for i in range(self.num_particles):
                lambda_u_i, zeta_u_i = lambdas_u_i[i], zetas_u_i[i]
                v_j = new_particles_vs[i][item, :]
                lambda_u_i += 1 / self.var * (v_j @ v_j.T)
                zeta_u_i += reward * v_j

                inv_lambda_u_i = np.linalg.inv(lambda_u_i)
                sampled_user_vector = np.random.multivariate_normal(
                    1 / self.var * (inv_lambda_u_i @ zeta_u_i), inv_lambda_u_i
                )
                new_particles_us[i][uid] = sampled_user_vector

                u_i = new_particles_us[i][self.items_consumed_users[item], :]
                lambda_v_i = 1 / self.var * (u_i.T @ u_i) + 1 / new_particles_var_vs[
                    i
                ] * np.eye(self.num_lat)

                zeta = np.sum(
                    np.multiply(
                        u_i,
                        np.array(self.items_consumed_users_rewards[item]).reshape(
                            -1, 1
                        ),
                    ),
                    axis=0,
                )
                inv_lambda_v_i = np.linalg.inv(lambda_v_i)
                item_sample_vector = np.random.multivariate_normal(
                    1 / self.var * (inv_lambda_v_i @ zeta), inv_lambda_v_i
                )
                new_particles_vs[i][item] = item_sample_vector

            if not updated_history:
                self.users_consumed_items[uid].append(item)
                self.users_consumed_items_rewards[uid].append(reward)
                self.items_consumed_users[item].append(uid)
                self.items_consumed_users_rewards[item].append(reward)
                updated_history = True

            self.particles_us = new_particles_us
            self.particles_vs = new_particles_vs
            self.particles_var_us = new_particles_var_us
            self.particles_var_vs = new_particles_var_vs


class PTSInit(PTS):
    def __init__(self, init, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init = init

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)

        if self.init == "entropy":
            items_entropy = value_functions.Entropy.get_items_entropy(
                self.train_consumption_matrix
            )
            self.items_bias = items_entropy
        elif self.init == "logpopent":
            items_entropy = value_functions.Entropy.get_items_entropy(
                self.train_consumption_matrix
            )
            items_popularity = value_functions.MostPopular.get_items_popularity(
                self.train_consumption_matrix, normalize=False
            )
            self.items_bias = value_functions.LogPopEnt.get_items_logpopent(
                items_popularity, items_entropy
            )

        self.items_bias = self.items_bias - np.min(self.items_bias)
        self.items_bias = self.items_bias / np.max(self.items_bias)
        assert self.items_bias.min() >= 0 and np.isclose(self.items_bias.max(), 1)
        res = scipy.optimize.minimize(
            lambda x, items_means, items_bias: np.sum(
                (items_bias - x @ items_means.T) ** 2
            ),
            np.ones(self.num_lat),
            args=(self.particles_vs[0], self.items_bias),
            method="BFGS",
        )
        self.initial_user_factors = res.x
        # print(np.corrcoef(self.items_bias,self.initial_user_factors @ self.items_means.T)[0,1])
        # self.bs = defaultdict(lambda: self.initial_b.copy())
        not_train_users = set(list(range(self.train_dataset.num_total_users))) - {
            int(self.train_dataset.data[i, 0])
            for i in range(len(self.train_dataset.data))
        }
        for i in range(self.num_particles):
            for uid in not_train_users:
                self.particles_us[i][uid] = self.initial_user_factors

            # self.particles_vs[i]


class PTSLogPopEnt(PTSInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init="logpopent", *args, **kwargs)



from . import dataset
import random
import numpy as np


class DatasetLoader:
    pass

class DefaultDatasetLoader:
    def __init__(
        self, dataset_processor, dataset_path, crono, random_seed, test_consumes, train_size
    ) -> None:
        self.dataset_processor = dataset_processor
        self.dataset_path = dataset_path
        self.crono = crono
        self.random_seed = random_seed
        self.test_consumes = test_consumes
        self.train_size = train_size
        self.dataset_processor = dataset_processor

    def load(self):
        np.random.seed(self.random_seed)

        data = self.dataset_processor.process(self.dataset_path)

        traintest_processor = dataset.TrainTestConsumption(
            crono=self.crono,
            test_consumes=self.test_consumes,
            train_size=self.train_size,
        )
        res = traintest_processor.process(data)
        return res

class DefaultValidationDatasetLoader:
    def __init__(
        self, dataset_processor, dataset_path, crono, random_seed, test_consumes, train_size
    ) -> None:
        self.dataset_processor = dataset_processor
        self.dataset_path = dataset_path
        self.crono = crono
        self.random_seed = random_seed
        self.test_consumes = test_consumes
        self.train_size = train_size

    def load(self):
        np.random.seed(self.random_seed)

        data = self.dataset_processor.process(self.dataset_path)
        traintest_processor = dataset.TrainTestConsumption(
            crono=self.crono,
            test_consumes=self.test_consumes,
            train_size=self.train_size,
        )
        res = traintest_processor.process(traintest_processor.process(data).train)

        return res

class ML100kDatasetLoader:
    def __init__(
        self, dataset_path, crono, random_seed, test_consumes, train_size
    ) -> None:
        self.dataset_path = dataset_path
        self.crono = crono
        self.random_seed = random_seed
        self.test_consumes = test_consumes
        self.train_size = train_size

    def load(self):
        np.random.seed(self.random_seed)
        random.seed(self.random_seed)
        ml100k_processor = dataset.MovieLens100k()
        data = ml100k_processor.process(self.dataset_path)

        traintest_processor = dataset.TrainTestConsumption(
            crono=self.crono,
            test_consumes=self.test_consumes,
            train_size=self.train_size,
        )
        res = traintest_processor.process(data)
        return res

class TRTEDatasetLoader:
    def __init__(
        self,
        dataset_path
        # self, dataset_path, crono, random_seed, test_consumes, train_size
    ) -> None:
        self.dataset_path = dataset_path
        # self.crono = crono
        # self.random_seed = random_seed
        # self.test_consumes = test_consumes
        # self.train_size = train_size

    def load(self):
        # np.random.seed(self.random_seed)
        trte_processor = dataset.TRTE()
        data = trte_processor.process(self.dataset_path)
        res = data
        # traintest_processor = dataset.TrainTestConsumption(
        # crono=self.crono,
        # test_consumes=self.test_consumes,
        # train_size=self.train_size,
        # )
        # res = traintest_processor.process(data)
        return res


class TRTEValidationDatasetLoader:
    def __init__(
        self, dataset_path, crono, random_seed, test_consumes, train_size
    ) -> None:
        self.dataset_path = dataset_path
        self.crono = crono
        self.random_seed = random_seed
        self.test_consumes = test_consumes
        self.train_size = train_size

    def load(self):
        np.random.seed(self.random_seed)
        random.seed(self.random_seed)
        trte_processor = dataset.TRTE()
        data = trte_processor.process(self.dataset_path)

        traintest_processor = dataset.TrainTestConsumption(
            crono=self.crono,
            test_consumes=self.test_consumes,
            train_size=self.train_size,
        )
        res = traintest_processor.process(data.train)

        return res
